import gzip
import importlib.util
import io
import json
import lzma
import os
import platform
import shutil
import signal
import socket
import subprocess
import sys
import tempfile
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple

from osbuild.util import host, linux


def find_kernel_dir(tree):
    modules_dir = Path(tree) / "usr" / "lib" / "modules"

    if not modules_dir.exists() or not modules_dir.is_dir():
        return None, None

    try:
        subdirs = [d for d in modules_dir.iterdir() if d.is_dir()]
    except PermissionError:
        return None, None

    for subdir in sorted(subdirs):
        vmlinuz_path = subdir / "vmlinuz"
        if vmlinuz_path.is_file():
            return modules_dir / subdir.name

    raise ValueError("No valid kernel directory found in /usr/lib/modules")


def get_module_dependencies(module_path):
    result = subprocess.run(
        ["modinfo", "-F", "depends", str(module_path)],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        return []

    depends = result.stdout.strip()
    if not depends:
        return []

    deps = [dep.strip() for dep in depends.split(",") if dep.strip()]
    return deps


class InitrdBuilder:
    def __init__(self, source_modules):
        self.tmpdir = None
        self.root = None
        self.source_modules = source_modules

        self.copied_modules = set()
        self.module_counter = 0

    def __enter__(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.root = Path(self.tmpdir.name)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tmpdir.cleanup()

    def mkdir(self, path):
        (self.root / path).mkdir(parents=True)

    def add_file(self, src, dest):
        shutil.copy2(src, self.root / dest)

    def add_binary(self, src, dest):
        self.add_file(src, dest)
        (self.root / dest).chmod(0o755)

    def find_module(self, module_name):
        patterns = [
            f"{module_name}.ko",
            f"{module_name}.ko.xz",
            f"{module_name}.ko.zst",
            f"{module_name}.ko.gz",
        ]

        for pattern in patterns:
            matches = list(self.source_modules.rglob(pattern))
            if matches:
                return matches[0]

        return None

    def decompress_module(self, module_path, dest_file):
        """Decompress a module file to destination."""
        module_name = module_path.name

        if module_name.endswith(".ko.xz"):
            with lzma.open(module_path, "rb") as f_in:
                with open(dest_file, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
        elif module_name.endswith(".ko.zst"):
            subprocess.run(
                ["zstd", "-dc", str(module_path)],
                stdout=open(dest_file, "wb"),
                check=True,
            )
        elif module_name.endswith(".ko.gz"):
            with gzip.open(module_path, "rb") as f_in:
                with open(dest_file, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
        else:
            shutil.copy2(module_path, dest_file)

    def copy_module(self, module_name):
        if module_name in self.copied_modules:
            return
        self.copied_modules.add(module_name)

        module_path = self.find_module(module_name)
        if not module_path:
            print(f"Warning: {module_name} module not found in {self.source_modules}", file=sys.stderr)
            return

        dependencies = get_module_dependencies(module_path)
        for dep in dependencies:
            self.copy_module(dep)

        modules_dir = self.root / "usr/lib/modules"

        if self.module_counter == 0:
            modules_dir.mkdir(parents=True)
        self.module_counter += 1

        prefix = f"{self.module_counter:03d}"
        dest_file = modules_dir / f"{prefix}-{module_name}.ko"

        self.decompress_module(module_path, dest_file)

    def write(self, dest):
        with open(dest, "wb") as f_out:
            result = subprocess.run(
                ["find", ".", "-print0"],
                cwd=self.root,
                stdout=subprocess.PIPE,
                check=True,
            )
            subprocess.run(
                ["cpio", "--null", "--create", "--format=newc", "--quiet"],
                cwd=self.root,
                input=result.stdout,
                stdout=f_out,
                check=True,
            )


def create_initrd(libdir, rootfs_dir, dest, extra_modules=None):
    kernel_dir = find_kernel_dir(rootfs_dir)

    with InitrdBuilder(kernel_dir) as builder:
        builder.mkdir("bin")
        builder.add_binary(Path(libdir) / "initrd/initrd", "init")
        builder.copy_module("virtiofs")
        if extra_modules:
            for module in extra_modules:
                print(f"Copying additional module: {module}")
                builder.copy_module(module)

        builder.write(dest)


def find_virtiofsd():
    binary_name = "virtiofsd"
    bin_dirs = ["/usr/libexec", "/usr/lib", "/usr/lib/qemu"]
    if "PATH" in os.environ:
        bin_dirs += os.environ["PATH"].split(":")

    for d in bin_dirs:
        p = os.path.join(d, binary_name)
        if os.path.isfile(p):
            return p

    raise RuntimeError("Can't find virtiofsd binary")


class Virtiofsd:
    def __init__(
        self,
        share_dir: str,
        socket_path: str,
        readonly: bool,
    ):
        self.share_dir = share_dir
        self.cmd = [
            find_virtiofsd(),
            "--socket-path=" + socket_path,
            "--shared-dir", share_dir,
            "--cache", "always",
            "--sandbox=none",
            "--xattr",
            "--security-label",
            "--log-level=off",
        ]
        if readonly:
            self.cmd += ["--readonly"]
        self._proc: Optional[subprocess.Popen] = None

    def start(self) -> int:
        if self._proc is not None:
            return self._proc.pid

        self._proc = subprocess.Popen(
            self.cmd,
            text=False,
            close_fds=True,
        )

        return self._proc.pid

    def stop(self) -> None:
        p = self._proc
        if not p:
            return

        try:
            p.terminate()
        except ProcessLookupError:
            self._proc = None
            return

        # Wait a bit for graceful exit
        try:
            p.wait(timeout=5.0)
        except subprocess.TimeoutExpired:
            try:
                p.kill()
            except ProcessLookupError:
                pass
            try:
                p.wait(timeout=2.0)
            except subprocess.TimeoutExpired:
                pass

        self._proc = None

    # Context manager
    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc, tb):
        self.stop()
        return False


def find_qemu(arch):
    binary_names = [f"qemu-system-{arch}", "qemu-kvm"]

    for binary_name in binary_names:
        qemu_bin_dirs = ["/usr/bin", "/usr/libexec"]
        if "PATH" in os.environ:
            qemu_bin_dirs += os.environ["PATH"].split(":")

        for d in qemu_bin_dirs:
            p = os.path.join(d, binary_name)
            if os.path.isfile(p):
                return p

    raise RuntimeError("Can't find qemu binary")


def qemu_available_accels(qemu):
    info = subprocess.check_output([qemu, "-accel", "help"]).decode('utf-8').splitlines()
    # First line is header, skip it
    return info[1:]


def set_pdeathsig():
    PR_SET_PDEATHSIG = 1
    libc = linux.Libc.default()
    libc.prctl(PR_SET_PDEATHSIG, signal.SIGTERM)


class Qemu:
    def __init__(
        self,
        mem: str,
        rootfs_path: str,
        libdir_path: str,
        serial_stdout: bool = False,
    ) -> None:
        self._pid: Optional[int] = None

        self._tmpdir = tempfile.TemporaryDirectory(prefix="osbuild-qemu-")
        self.pidfile = os.path.join(self._tmpdir.name, "qemu.pid")
        self.serials: Dict[str, str] = {}
        self.virtiofs: Dict[str, Tuple[str, Virtiofsd]] = {}
        self._proc: Optional[subprocess.Popen] = None

        kerneldir = find_kernel_dir(rootfs_path)
        kernel_path = kerneldir / "vmlinuz"

        initrd_path = Path(self._tmpdir.name) / "initrd"
        create_initrd(libdir_path, rootfs_path, initrd_path)

        self.qmp_path = Path(self._tmpdir.name) / "qmp.sock"

        arch = platform.machine()
        qemu_bin = find_qemu(arch)
        qemu_accels = qemu_available_accels(qemu_bin)

        self.cmd = [
            qemu_bin,
            "-qmp", f"unix:{self.qmp_path},server=on,wait=off",
            "-display",
            "none",
            "-m",
            mem,
            # This is needed for virtiofs, and size must match -m
            "-object",
            f"memory-backend-memfd,id=mem0,size={mem},share=on",
            "-numa",
            "node,memdev=mem0",
        ]
        self._id_counter = 0

        console = "ttyS0"
        cpu = None
        if arch == "aarch64":
            console = "ttyAMA0"
            self.cmd += ["-machine", "virt"]
            cpu = "cortex-a57"
        elif arch == "x86_64":
            self.cmd += ["-machine", "q35"]
            cpu = "Haswell"  # x86-64 v3 - required by cs10

        if serial_stdout:
            self.cmd += ["-serial", "file:/dev/stdout"]

        if "kvm" in qemu_accels and os.path.exists("/dev/kvm"):
            self.cmd += ["-enable-kvm"]
            cpu = "host"  # Override arch-specific cpu

        if cpu is not None:
            self.cmd += ["-cpu", cpu]

        init = "/run/mnt/mnt0/osbuild/vm.py"
        cmdline = f"console={console} quiet selinux=1 enforcing=0 rootfstype=virtiofs root=rootfs ro mount=mnt0 init={init}"

        # vm.py will add its parent to the search path to find the "osbuild" module, so
        # mount the directory with the osbuild directory at /mnt and run /mnt/osbuild/vm.py
        spec = importlib.util.find_spec("osbuild")
        assert spec is not None and spec.origin is not None
        modpath = os.path.dirname(spec.origin)  # $some_python_path/osbuild
        modpath = os.path.dirname(modpath)  # $some_python_path

        self.add_kernel(str(kernel_path), str(initrd_path), cmdline)
        self.add_virtio_serial("ipc.0")
        self.add_virtiofs(rootfs_path, "rootfs", readonly=True)
        self.add_virtiofs(modpath, "mnt0", readonly=True)
        self.add_virtiofs(libdir_path, "libdir", readonly=True)
        try:
            container_storage_conf = host.get_container_storage()
            container_storage = container_storage_conf["storage"].get("graphroot", "/var/lib/containers/storage")
            if Path(container_storage).is_dir():
                self.add_virtiofs(container_storage, "containers", readonly=True)
        except FileNotFoundError:
            pass

        self.ipc = None

    def add_arguments(self, args: List[str]) -> None:
        self.cmd += args

    def add_kernel(self, kernel_path: str, initrd_path: str, commandline: str):
        self.add_arguments(
            ["-kernel", kernel_path, "-initrd", initrd_path, "-append", commandline]
        )

    def add_virtio_serial(self, name):
        if name in self.serials:
            raise RuntimeError(f"Virtio serial name {name} already used")
        socket_path = os.path.join(self._tmpdir.name, "serial-" + name + ".sock")
        self.serials[name] = socket_path
        char = self._alloc_chardev()
        self.add_arguments(
            [
                "-device",
                "virtio-serial",
                "-chardev",
                f"socket,path={socket_path},server=on,wait=off,id={char}",
                "-device",
                f"virtserialport,chardev={char},name={name}",
            ]
        )

    def connect_virtio_serial(self, name):
        if name not in self.serials:
            raise RuntimeError(f"Virtio serial name {name} doesn't exist")

        client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        client.connect(self.serials[name])
        raw = client.makefile("rwb", buffering=0)
        txt = io.TextIOWrapper(raw, encoding="utf-8", newline="\n", line_buffering=True)

        return txt

    def ensure_ipc(self):
        if not self.ipc:
            self.ipc = self.connect_virtio_serial("ipc.0")

    def send_request(self, obj):
        self.ensure_ipc()
        self.ipc.write(json.dumps(obj, separators=(",", ":")))
        self.ipc.write("\n")
        self.ipc.flush()

    def read_response(self):
        line = self.ipc.readline()
        return json.loads(line)

    def add_virtiofs(self, share_path, tag, readonly):
        if tag in self.virtiofs:
            raise RuntimeError(f"Virtiofs {tag} already used")
        socket_path = os.path.join(self._tmpdir.name, "virtiofs-" + tag + ".sock")
        virtfsd = Virtiofsd(share_path, socket_path, readonly)
        self.virtiofs[tag] = (socket_path, virtfsd)
        char = self._alloc_chardev()
        self.add_arguments(
            [
                "-chardev",
                f"socket,id={char},path={socket_path}",
                "-device",
                f"vhost-user-fs-pci,queue-size=1024,chardev={char},tag={tag}",
            ]
        )

    def wait_for_qmp(self, timeout=10.0):
        deadline = time.time() + timeout
        while time.time() < deadline:
            try:
                s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                s.connect(str(self.qmp_path))
                s.close()
                return True
            except (FileNotFoundError, ConnectionRefusedError):
                time.sleep(0.1)
        return False

    def start(self) -> subprocess.Popen:
        for _path, virtfsd in self.virtiofs.values():
            virtfsd.start()

        # We try to properly clean up the child qemu in stop(), however
        # we also call pdeathsig to enure the child is killed if the parent
        # dies unexpectedly (sigkill for example)
        self._proc = subprocess.Popen(self.cmd, preexec_fn=set_pdeathsig)  # noqa: PLW1509 # pylint: disable=W1509

        # Wait until QMP responds, at that point we know the VM setup
        # is completed.
        if not self.wait_for_qmp():
            self._proc.terminate()
            self._proc = None
            raise RuntimeError("QEMU did not become ready")

        return self._proc

    def stop(self) -> None:
        """Terminate QEMU and clean up the pidfile we own."""
        proc = self._proc
        if proc is None or proc.poll() is not None:
            self.cleanup()
            return

        try:
            proc.terminate()
            proc.wait(timeout=5.0)
        except subprocess.TimeoutExpired:
            proc.kill()

        self.cleanup()

    def cleanup(self):
        for _path, virtfsd in self.virtiofs.values():
            virtfsd.stop()
        if self._tmpdir:
            self._tmpdir.cleanup()
            self._tmpdir = None

    # Context manager hooks
    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc, tb):
        self.stop()
        return False

    def __del__(self):
        # Best-effort cleanup if user forgets to exit/stop
        try:
            self.stop()
        except Exception:  # pylint: disable=W0718
            pass

    def monitored_request(self, monitor, opname, **kwargs):
        op = {"op": opname, **kwargs}
        self.send_request(op)
        while True:
            response = self.read_response()
            if "log" in response:
                monitor.log(response["log"])
                continue

            if "ok" not in response:
                raise RuntimeError(f"VM Operation {opname} Unexpected response: {str(response)}")

            ok = response["ok"]
            if not ok:
                errtype = response.get("error")

                if errtype == "handler_exception":
                    raise RuntimeError(
                        f"VM Operation {opname} raised exception: " +
                        response['msg'] + ", trace:\n" + response['trace'])

                raise RuntimeError(f"VM Operation {opname} failed: {response['msg']}")

            return response["r"]

    # --- Internals ------------------------------------------------------------
    def _alloc_chardev(self):
        devid = self._id_counter
        self._id_counter += 1
        return f"char{devid}"

    @staticmethod
    def _is_pid_alive(pid: int) -> bool:
        try:
            os.kill(pid, 0)
            return True
        except ProcessLookupError:
            return False
        except PermissionError:
            # Exists but not ours
            return True

    @staticmethod
    def _wait_for_exit(pid: int, timeout: float) -> bool:
        deadline = time.time() + timeout
        while time.time() < deadline:
            if not Qemu._is_pid_alive(pid):
                return True
            time.sleep(0.05)
        return False



#!/usr/bin/python3
from unittest import mock

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.vagrant"


# Prepare dataset containing good and bad API call parameters
@pytest.mark.parametrize("test_data, expected_err", [
    # Bad API parameters
    ({}, "not valid under any of the given schemas"),
    ({"provider": "none"}, "not valid under any of the given schemas"),
    ({"provider": "virtualbox"}, "not valid under any of the given schemas"),
    ({"provider": "virtualbox", "virtualbox": {}}, "not valid under any of the given schemas"),
    ({"provider": "libvirt", "virtualbox": {"mac_address": "1"}}, "not valid under any of the given schemas"),
    ({"provider": "libvirt", "synced_folders": {"/vagrant": {"type": "vboxfs"}}}, "not valid under any of the given schemas"),
    # Good API parameters
    ({"provider": "libvirt"}, ""),
    ({"provider": "virtualbox", "virtualbox": {"mac_address": "000000000000"},
      "synced_folders": {"/vagrant": {"type": "rsync"}}}, ""),
])
# This test validates only API calls using correct and incorrect queries
def test_schema_validation_vagrant(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "devices": {
            "device": {
                "path": "some-path",
            },
        },
        "options": {
        }
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False, f"err: {[e.as_dict() for e in res.errors]}"
        testutil.assert_jsonschema_error_contains(res, expected_err)


@mock.patch("subprocess.run")
def test_vagrant_writes_file_without_rsync_when_guest_additions(_mock_run, tmp_path, stage_module):
    treepath = tmp_path / "tree"
    treepath.mkdir()

    vagrantfile = treepath / "Vagrantfile"
    metadatafile = treepath / "metadata.json"

    options = {
        "provider": "virtualbox",
        "virtualbox": {
            "mac_address": "000000000000",
        },
        "synced_folders": {
            "/vagrant": {
                "type": "rsync",
            },
        }
    }

    inputs = {
        "image": {
            "path": "/foo",
            "data": {
                "files": {
                    "image.vmdk": {"path": "foo.vmdk"}
                }
            }
        }
    }

    stage_module.main(treepath, options, inputs)

    assert vagrantfile.exists()
    assert metadatafile.exists()

    assert vagrantfile.read_text() == """Vagrant.configure("2") do |config|
    config.vm.base_mac = "000000000000"
    config.vm.synced_folder ".", "/vagrant", type: "rsync"

end
"""

    assert metadatafile.read_text() == '{"provider": "virtualbox"}'


@mock.patch("subprocess.run")
@mock.patch("subprocess.check_output")
def test_vagrant_writes_file_for_libvirt(mock_check_output, _mock_run, tmp_path, stage_module):
    mock_check_output.return_value = '{"virtual-size": 1000000000}'

    treepath = tmp_path / "tree"
    treepath.mkdir()

    vagrantfile = treepath / "Vagrantfile"
    metadatafile = treepath / "metadata.json"

    options = {
        "provider": "libvirt",
    }

    inputs = {
        "image": {
            "path": "/foo",
            "data": {
                "files": {
                    "image.vmdk": {"path": "foo.vmdk"}
                }
            }
        }
    }

    stage_module.main(treepath, options, inputs)

    assert vagrantfile.exists()
    assert metadatafile.exists()

    assert vagrantfile.read_text() == """Vagrant.configure("2") do |config|
    config.vm.provider :libvirt do |libvirt|
  libvirt.driver = "kvm"
end

end
"""

    assert metadatafile.read_text() == '{"provider": "libvirt", "format": "qcow2", "virtual_size": 1}'



import importlib
from enum import Enum
from types import ModuleType
from typing import Dict, TextIO

from osbuild.solver.exceptions import InvalidAPIVersionError
from osbuild.solver.model import DepsolveResult, DumpResult, SearchResult
from osbuild.solver.request import SolverRequest


class SolverAPIVersion(Enum):
    V1 = 1
    V2 = 2

    def __str__(self) -> str:
        return self.name.lower()


def get_api_module(api_version: SolverAPIVersion) -> ModuleType:
    try:
        return importlib.import_module(f"osbuild.solver.api.v{api_version.value}")
    except ImportError as e:
        raise InvalidAPIVersionError(f"Failed to import solver API module for version: {api_version.value}") from e


def parse_request(request_dict: Dict) -> SolverRequest:
    """
    Parse a request dict into a SolverRequest object

    Detects API version from the request (defaults to V1) and
    uses the appropriate version-specific parser.
    """
    # Get API version from the request, default to V1, because the current (V1)
    # API implementation does not set any version field.
    # pylint: disable=fixme
    # XXX: remove the default and make it an error once V1 is deprecated.
    version_num = request_dict.get("api_version", 1)
    try:
        api_version = SolverAPIVersion(version_num)
    except ValueError as e:
        raise InvalidAPIVersionError(f"Invalid API version: {e}") from e

    api_module = get_api_module(api_version)
    return api_module.parse_request(request_dict)


def serialize_response_depsolve(
        api_version: SolverAPIVersion, solver: str, result: DepsolveResult, writer: TextIO) -> None:
    """
    Serializes a Solver API response for the DEPSOLVE command.

    NOTE: The actual types returned by the API version specific implementations
          of these functions will differ across implementations. It does not
          make sense to define it in the main proxy serializers.
    """
    api_module = get_api_module(api_version)
    api_module.serialize_response_depsolve(solver, result, writer)


def serialize_response_dump(
        api_version: SolverAPIVersion, solver: str, result: DumpResult, writer: TextIO) -> None:
    """
    Serializes a Solver API response for the DUMP command.

    NOTE: The actual types returned by the API version specific implementations
          of these functions will differ across implementations. It does not
          make sense to define it in the main proxy serializers.
    """
    api_module = get_api_module(api_version)
    api_module.serialize_response_dump(solver, result, writer)


def serialize_response_search(
        api_version: SolverAPIVersion, solver: str, result: SearchResult, writer: TextIO) -> None:
    """
    Serializes a Solver API response for the SEARCH command.

    NOTE: The actual types returned by the API version specific implementations
          of these functions will differ across implementations. It does not
          make sense to define it in the main proxy serializers.
    """
    api_module = get_api_module(api_version)
    api_module.serialize_response_search(solver, result, writer)



#!/usr/bin/python3

import os
import subprocess

import pytest

from osbuild.testutil import has_executable, make_container

SOURCES_NAME = "org.osbuild.containers-storage"


@pytest.mark.skipif(not has_executable("podman"), reason="no podman executable")
@pytest.mark.skipif(os.getuid() != 0, reason="root only")
def test_containers_storage_integration(tmp_path, sources_service):
    with make_container(tmp_path, {
        "file1": "file1 content",
    }) as base_tag:
        image_id = subprocess.check_output(["podman", "inspect", "-f", "{{ .Id }}", base_tag],
                                           universal_newlines=True).strip()
        checksum = f"sha256:{image_id}"
        assert sources_service.exists(checksum, None)


@pytest.mark.skipif(not has_executable("podman"), reason="no podman executable")
@pytest.mark.skipif(os.getuid() != 0, reason="root only")
def test_containers_storage_integration_missing(sources_service):
    checksum = "sha256:1234567890123456789012345678901234567890909b14ffb032aa20fa23d9ad6"
    assert not sources_service.exists(checksum, None)


@pytest.mark.skipif(not has_executable("podman"), reason="no podman executable")
@pytest.mark.skipif(os.getuid() != 0, reason="root only")
def test_containers_storage_integration_invalid(sources_service):
    # put an invalid reference into the source to ensure skopeo errors with
    # a different error than image not found
    checksum = "sha256:["
    with pytest.raises(RuntimeError) as exc:
        sources_service.exists(checksum, None)
    assert "unknown skopeo error:" in str(exc)



"""
Request model for the solver API

This module contains data classes representing solver requests and their
components. These models are populated from API request dicts by version-
specific parsers in the api/ module.
"""

from enum import Enum
from typing import TYPE_CHECKING, List, Optional

from osbuild.solver.exceptions import InvalidRequestError
from osbuild.solver.model import Repository

if TYPE_CHECKING:
    from osbuild.solver.api import SolverAPIVersion


class SolverCommand(Enum):
    """Supported solver commands"""
    DEPSOLVE = "depsolve"
    DUMP = "dump"
    SEARCH = "search"

    def __str__(self) -> str:
        return self.value


class DepsolveTransaction:
    """A single transaction for dependency resolution"""

    def __init__(
        self,
        package_specs: List[str],
        exclude_specs: Optional[List[str]] = None,
        repo_ids: Optional[List[str]] = None,
        module_enable_specs: Optional[List[str]] = None,
        install_weak_deps: bool = False,
    ):
        # pylint: disable=fixme
        # XXX: We can't enforce this, because there is a "bug" in the osbuild/images "os" pipeline implementation.
        # https://github.com/osbuild/image-builder/commit/8b779619aa0c3a9b8537f6bb79324303cb87909c introduced three
        # transactions, but the "customizations" package set is added to the "chain" unconditionally, even if it
        # is empty. This is the case for "container" and "wsl" image types. Other image types enable at least
        # SELinux, which adds the "selinux-policy-targeted" package to the list.
        # if not package_specs:
        #     raise InvalidRequestError("Depsolve transaction must contain at least one package specification")

        self.package_specs = package_specs
        self.exclude_specs = exclude_specs
        self.repo_ids = repo_ids
        self.module_enable_specs = module_enable_specs
        self.install_weak_deps = install_weak_deps

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DepsolveTransaction):
            return False
        return (
            self.package_specs == other.package_specs
            and self.exclude_specs == other.exclude_specs
            and self.repo_ids == other.repo_ids
            and self.module_enable_specs == other.module_enable_specs
            and self.install_weak_deps == other.install_weak_deps
        )

    def __hash__(self) -> int:
        return hash((
            tuple(self.package_specs),
            tuple(self.exclude_specs) if self.exclude_specs else None,
            tuple(self.repo_ids) if self.repo_ids else None,
            tuple(self.module_enable_specs) if self.module_enable_specs else None,
            self.install_weak_deps,
        ))

    def __repr__(self) -> str:
        return f"DepsolveTransaction(package_specs={self.package_specs}, exclude_specs={self.exclude_specs}, " \
            f"repo_ids={self.repo_ids}, module_enable_specs={self.module_enable_specs}, " \
            f"install_weak_deps={self.install_weak_deps})"


class SBOMRequest:
    """
    SBOM generation request

    XXX: this will be deprecated once the SBOM generation is moved to osbuild/images
    """

    def __init__(self, sbom_type: str):
        if not sbom_type:
            raise InvalidRequestError("SBOM type cannot be empty")
        if sbom_type != "spdx":
            raise InvalidRequestError(f"Unsupported SBOM type '{sbom_type}'. Supported types: spdx")
        self.sbom_type = sbom_type

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SBOMRequest):
            return False
        return self.sbom_type == other.sbom_type

    def __hash__(self) -> int:
        return hash(self.sbom_type)


class DepsolveCmdArgs:
    """Arguments for the DEPSOLVE command"""

    def __init__(self, transactions: List[DepsolveTransaction], sbom_request: Optional[SBOMRequest] = None):
        if not transactions:
            raise InvalidRequestError("Depsolve command must contain at least one transaction")

        self.transactions = transactions
        self.sbom_request = sbom_request

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, DepsolveCmdArgs):
            return False
        return self.transactions == other.transactions and self.sbom_request == other.sbom_request

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.transactions),
                self.sbom_request,
            )
        )

    def __repr__(self) -> str:
        return f"DepsolveCmdArgs(transactions={self.transactions}, sbom_request={self.sbom_request})"


class SearchCmdArgs:
    """Arguments for the SEARCH command"""

    def __init__(self, packages: List[str], latest: bool = False):
        self.packages = packages
        self.latest = latest

        if not packages:
            raise InvalidRequestError("Search command must contain at least one package specification")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SearchCmdArgs):
            return False
        return self.packages == other.packages and self.latest == other.latest

    def __hash__(self) -> int:
        return hash((tuple(self.packages), self.latest))


class SolverConfig:
    """Solver configuration"""

    def __init__(
        self,
        arch: str,
        cachedir: str,
        releasever: Optional[str] = None,
        module_platform_id: Optional[str] = None,
        proxy: Optional[str] = None,
        repos: Optional[List[Repository]] = None,
        root_dir: Optional[str] = None,
        optional_metadata: Optional[List[str]] = None,
    ):
        self.arch = arch
        self.releasever = releasever
        self.cachedir = cachedir
        self.module_platform_id = module_platform_id
        self.proxy = proxy
        self.repos = repos
        self.root_dir = root_dir
        self.optional_metadata = optional_metadata

        # Validation
        required_args = ["arch", "cachedir"]
        for arg in required_args:
            if not getattr(self, arg):
                raise InvalidRequestError(f"Field '{arg}' is required")

        # Repository validation
        if not repos and not root_dir:
            raise InvalidRequestError("No 'repos' or 'root_dir' specified")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SolverConfig):
            return False
        return (
            self.arch == other.arch
            and self.releasever == other.releasever
            and self.cachedir == other.cachedir
            and self.module_platform_id == other.module_platform_id
            and self.proxy == other.proxy
            and self.repos == other.repos
            and self.root_dir == other.root_dir
            and self.optional_metadata == other.optional_metadata
        )

    def __hash__(self) -> int:
        return hash((
            self.arch,
            self.releasever,
            self.cachedir,
            self.module_platform_id,
            self.proxy,
            tuple(self.repos) if self.repos else None,
            self.root_dir,
            tuple(self.optional_metadata) if self.optional_metadata else None,
        ))

    def __repr__(self) -> str:
        return f"SolverConfig(arch={self.arch}, releasever={self.releasever}, cachedir={self.cachedir}, " \
            f"module_platform_id={self.module_platform_id}, proxy={self.proxy}, repos={self.repos}, " \
            f"root_dir={self.root_dir}, optional_metadata={self.optional_metadata})"


# pylint: disable=too-many-instance-attributes
class SolverRequest:
    """Version-agnostic solver request"""

    def __init__(
        self,
        api_version: "SolverAPIVersion",
        command: SolverCommand,
        config: SolverConfig,
        # Command-specific args
        depsolve_args: Optional[DepsolveCmdArgs] = None,
        search_args: Optional[SearchCmdArgs] = None,
    ):
        self.api_version = api_version
        self.command = command
        self.config = config
        self.search_args = search_args
        self.depsolve_args = depsolve_args

        if not command:
            raise InvalidRequestError("Field 'command' is required")

        if command not in [SolverCommand.DEPSOLVE, SolverCommand.SEARCH, SolverCommand.DUMP]:
            raise InvalidRequestError(
                f"Invalid command '{command}': must be one of {', '.join([c.value for c in SolverCommand])}")

        # Command-specific validation
        if command != SolverCommand.DEPSOLVE and depsolve_args:
            raise InvalidRequestError("Depsolve arguments are only supported with 'depsolve' command")

        if command != SolverCommand.SEARCH and search_args:
            raise InvalidRequestError("Search arguments are only supported with 'search' command")

        if command == SolverCommand.DEPSOLVE and not depsolve_args:
            raise InvalidRequestError("Depsolve command requires arguments")

        if command == SolverCommand.SEARCH and not search_args:
            raise InvalidRequestError("Search command requires arguments")

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SolverRequest):
            return False
        return (
            self.api_version == other.api_version
            and self.command == other.command
            and self.config == other.config
            and self.depsolve_args == other.depsolve_args
            and self.search_args == other.search_args
        )

    def __hash__(self) -> int:
        return hash((
            self.api_version,
            self.command,
            self.config,
            self.depsolve_args,
            self.search_args,
        ))

    def __repr__(self) -> str:
        return f"SolverRequest(api_version={self.api_version}, command={self.command}, config={self.config}, " \
            f"depsolve_args={self.depsolve_args}, search_args={self.search_args})"



#!/usr/bin/python3
import os
import struct

import pytest

STAGE_NAME = "org.osbuild.createaddrsize"


def test_createaddrsize(tmp_path, stage_module):
    # make a fake initrd.img with a few bytes in it
    with open(f"{tmp_path}/initrd.img", "wb") as f:
        f.write(b"A file pretending to be an initrd.img for testing purposes")

    options = {
        "initrd": "initrd.img",
        "addrsize": "initrd.addrsize"
    }

    stage_module.main(tmp_path, options)
    assert os.path.exists(f"{tmp_path}/initrd.addrsize")
    with open(f"{tmp_path}/initrd.addrsize", "rb") as f:
        data = f.read()
    assert struct.unpack(">iiii", data) == (0, 0x2000000, 0, 58)


@pytest.mark.parametrize("test_data,expected_err", [
    # good
    (
        {
            "initrd": "/images/initrd.img",
            "addrsize": "/images/initrd.addrsize"
        },
        ""
    ),
    (
        {
            "initrd": "images/initrd.img",
            "addrsize": "images/initrd.addrsize"
        },
        ""
    ),
    # bad
    (
        {}, ["'addrsize' is a required property",
             "'initrd' is a required property"]
    ),
    (
        {
            "initrd": "/images/initrd.img",
        },
        ["'addrsize' is a required property"]
    ),
    (
        {
            "initrd": "../images/initrd.img",
            "addrsize": "images/initrd.addrsize"
        },
        ["'../images/initrd.img' does not match '^\\\\/?(?!\\\\.\\\\.)((?!\\\\/\\\\.\\\\.\\\\/).)+$'"]
    ),
])
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {
        "type": STAGE_NAME,
        "options": {},
    }
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        err_msgs = sorted([e.as_dict()["message"] for e in res.errors])
        assert err_msgs == expected_err



import abc
import os
import urllib.error
import urllib.parse
import urllib.request
from typing import Optional

from osbuild.solver.exceptions import GPGKeyReadError, InvalidRequestError, NoRHSMSubscriptionsError
from osbuild.solver.model import DepsolveResult, DumpResult, Repository, SearchResult
from osbuild.solver.request import DepsolveCmdArgs, SearchCmdArgs, SolverConfig
from osbuild.util.rhsm import Subscriptions


class Solver(abc.ABC):
    @abc.abstractmethod
    def dump(self) -> DumpResult:
        pass

    @abc.abstractmethod
    def depsolve(self, args: DepsolveCmdArgs) -> DepsolveResult:
        pass

    @abc.abstractmethod
    def search(self, args: SearchCmdArgs) -> SearchResult:
        pass


class SolverBase(Solver):
    # put any shared helpers in here

    # Override this in the subclass
    SOLVER_NAME = "unknown"

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.SOLVER_NAME == SolverBase.SOLVER_NAME:
            raise ValueError(f"{cls.__name__} must override SOLVER_NAME")

    def __init__(
        self,
        config: SolverConfig,
        persistdir: os.PathLike,
        license_index_path: Optional[os.PathLike] = None,
    ):
        self.config = config
        self.persistdir = persistdir
        self.license_index_path = license_index_path
        # Set of repository IDs that need RHSM secrets
        self.repo_ids_with_rhsm = set()

        # Get the RHSM secrets for the repositories that need them
        subscriptions = None
        for repo in self.config.repos or []:
            if not repo.rhsm:
                continue

            repo_urls = []
            if repo.baseurl:
                repo_urls.extend(repo.baseurl)
            if repo.metalink:
                repo_urls.append(repo.metalink)
            if repo.mirrorlist:
                repo_urls.append(repo.mirrorlist)

            if subscriptions is None:
                try:
                    subscriptions = Subscriptions.from_host_system()
                except RuntimeError as e:
                    raise NoRHSMSubscriptionsError("The host system does not have any valid subscriptions. "
                                                   "Subscribe it before specifying rhsm: true in repositories "
                                                   f"(error details: {e}; repo_id: {repo.repo_id}; "
                                                   f"repo_urls: {repo_urls})") from e

            # We will override the sslcacert, sslclientkey, and sslclientcert with the ones from the subscriptions
            # Return an error if the fields are already set
            if repo.sslcacert or repo.sslclientkey or repo.sslclientcert:
                raise InvalidRequestError("The sslcacert, sslclientkey, and sslclientcert fields cannot be set "
                                          "when rhsm: true is specified")

            try:
                secrets = subscriptions.get_secrets(repo_urls)
                repo.sslcacert = secrets["ssl_ca_cert"]
                repo.sslclientkey = secrets["ssl_client_key"]
                repo.sslclientcert = secrets["ssl_client_cert"]
            except RuntimeError as e:
                raise NoRHSMSubscriptionsError(f"Error getting RHSM secrets for {repo_urls}: {e}") from e

            self.repo_ids_with_rhsm.add(repo.repo_id)

    def set_rhsm_flag(self, repo: Repository) -> None:
        """
        Set the rhsm flag on the repository based on repo_ids_with_rhsm.
        Sets the flag to True if repo.repo_id is in the set of RHSM repos, False otherwise.
        Otherwise, set it to False.
        """
        repo.rhsm = repo.repo_id in self.repo_ids_with_rhsm


def modify_rootdir_path(path, root_dir):
    if path and root_dir:
        # if the root_dir is set, we need to translate the key path to be under this directory
        return os.path.join(root_dir, path.lstrip("/"))
    return path


def read_keys(paths, root_dir=None):
    keys = []
    for path in paths:
        url = urllib.parse.urlparse(path)
        if url.scheme == "file":
            path = url.path
            path = modify_rootdir_path(path, root_dir)
            try:
                with open(path, mode="r", encoding="utf-8") as keyfile:
                    keys.append(keyfile.read())
            except Exception as e:
                raise GPGKeyReadError(f"error loading gpg key from {path}: {e}") from e
        elif url.scheme in ["http", "https"]:
            try:
                resp = urllib.request.urlopen(urllib.request.Request(path))
                keys.append(resp.read().decode())
            except urllib.error.URLError as e:
                raise GPGKeyReadError(f"error reading remote gpg key at {path}: {e}") from e
        else:
            raise GPGKeyReadError(f"unknown url scheme for gpg key: {url.scheme} ({path})")
    return keys



"""
Monitor pipeline activity

The osbuild `Pipeline` class supports monitoring of its activities
by providing a monitor object that implements the `BaseMonitor`
interface. During the execution of the pipeline various functions
are called on the monitor object at certain events. Consult the
`BaseMonitor` class for the description of all available events.
"""

import abc
import copy
import datetime
import hashlib
import json
import os
import sys
import time
from threading import Lock
from typing import Any, Dict, Optional, Set, Union

import osbuild
from osbuild.pipeline import StageResult
from osbuild.sources import SourceResults
from osbuild.util.term import fmt as vt


def omitempty(d: dict):
    """ Omit None and empty string ("") values from the given dict """
    for k, v in list(d.items()):
        if v is None or v == "":
            del d[k]
        elif isinstance(v, dict):
            omitempty(v)
    return d


class Context:
    """Context for a single log entry. Automatically calculates hash/id when read."""

    def __init__(self,
                 origin: Optional[str] = None,
                 pipeline: Optional[osbuild.Pipeline] = None,
                 stage: Optional[osbuild.Stage] = None):
        self._origin = origin
        self._pipeline_name = pipeline.name if pipeline else None
        self._pipeline_id = pipeline.id if pipeline else None
        self._stage_name = stage.name if stage else None
        self._stage_id = stage.id if stage else None
        self._id = None
        self._id_history: Set[str] = set()

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        # reset "_id" on any write so that the hash is automatically recalculated
        if name != "_id":
            super().__setattr__("_id", None)

    def with_origin(self, origin: Optional[str]) -> "Context":
        """
        Return a Context with the given origin but otherwise identical.

        Note that if the origin is empty or same it will return self.
        """
        if origin is None or origin == self._origin:
            return self
        ctx = copy.copy(self)
        ctx.origin = origin
        return ctx

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, origin: str):
        self._origin = origin

    @property
    def pipeline_name(self):
        return self._pipeline_name

    @property
    def pipeline_id(self):
        return self._pipeline_id

    def set_pipeline(self, pipeline: osbuild.Pipeline):
        self._pipeline_name = pipeline.name
        self._pipeline_id = pipeline.id

    @property
    def stage_name(self):
        return self._stage_name

    @property
    def stage_id(self):
        return self._stage_id

    def set_stage(self, stage: osbuild.Stage):
        self._stage_name = stage.name
        self._stage_id = stage.id

    @property
    def id(self):
        if self._id is None:
            self._id = hashlib.sha256(
                json.dumps(self._dict(), sort_keys=True).encode()).hexdigest()
        return self._id

    def _dict(self):
        return {
            "origin": self._origin,
            "pipeline": {
                "name": self._pipeline_name,
                "id": self._pipeline_id,
                "stage": {
                    "name": self._stage_name,
                    "id": self._stage_id,
                },
            },
        }

    def as_dict(self):
        d = self._dict()
        ctxid = self.id
        if ctxid in self._id_history:
            return {"id": self.id}
        d["id"] = self.id
        self._id_history.add(self.id)
        return d


class Progress:
    """Progress represents generic progress information.

    A progress can contain a sub_progress to track more
    nested progresses. Any increment of a parent progress
    will the reset the sub_progress to None and a new
    sub_progress needs to be provided.

    Keyword arguments:
    name  -- user visible name for the progress
    total -- total steps required to finish the progress
    """

    def __init__(self, name: str, total: int):
        self.name = name
        self.total = total
        self.done = 0
        self.sub_progress: Optional[Progress] = None

    def incr(self):
        """Increment the "done" count"""
        self.done += 1
        if self.sub_progress:
            self.sub_progress = None

    def as_dict(self):
        d = {
            "name": self.name,
            "total": self.total,
            "done": self.done,
        }
        if self.sub_progress:
            d["progress"] = self.sub_progress.as_dict()
        return d


def log_entry(message: Optional[str] = None,
              context: Optional[Context] = None,
              progress: Optional[Progress] = None,
              options: Optional[Dict] = None,
              duration: Optional[float] = None,
              result: Union[StageResult, SourceResults, None] = None,
              metadata: Optional[Dict] = None,
              ) -> dict:
    """
    Create a single log entry dict with a given message, context, and progress objects.
    All arguments are optional. A timestamp is added to the message.
    """
    # we probably want to add an (optional) error message here too once the
    # monitors support that
    return omitempty({
        "message": message,
        "result": result.as_dict() if result else None,
        "context": context.as_dict() if context else None,
        "progress": progress.as_dict() if progress else None,
        "options": options if options else None,
        "timestamp": time.time(),
        "duration": duration,
        "metadata": metadata if metadata else None,
    })


class TextWriter:
    """Helper class for writing text to file descriptors"""

    def __init__(self, fd: int):
        self.fd = fd
        self.isatty = os.isatty(fd)

    def term(self, text, *, clear=False):
        """Write text if attached to a terminal."""
        if not self.isatty:
            return

        if clear:
            self.write(vt.reset)

        self.write(text)

    def write(self, text: str):
        """Write all of text to the log file descriptor"""
        data = text.encode("utf-8")
        n = len(data)
        while n:
            k = os.write(self.fd, data)
            n -= k
            if n:
                data = data[n:]


class BaseMonitor(abc.ABC):
    """Base class for all pipeline monitors"""

    def __init__(self, fd: int, _: int = 0) -> None:
        """Logging will be done to file descriptor `fd`"""
        self.out = TextWriter(fd)

    def begin(self, pipeline: osbuild.Pipeline):
        """Called once at the beginning of a pipeline"""

    def finish(self, results: Dict):
        """Called at the very end of a pipeline"""

    def stage(self, stage: osbuild.Stage):
        """Called when a stage is being built"""

    def assembler(self, assembler: osbuild.Stage):
        """Called when an assembler is being built"""

    def result(self, result: Union[StageResult, SourceResults], metadata: Optional[Dict] = None) -> None:
        """Called when a module (stage/assembler) is done with its result"""

    # note that this should be re-entrant
    def log(self, message: str, origin: Optional[str] = None):
        """Called for all module log outputs"""


class NullMonitor(BaseMonitor):
    """Monitor class that does not report anything"""


class LogMonitor(BaseMonitor):
    """Monitor that follows the log output of modules

    This monitor will print a header with `name: id` followed
    by the options for each module as it is being built. The
    full log messages of the modules will be printed as soon as
    they become available.
    The constructor argument `fd` is a file descriptor, where
    the log will get written to. If `fd`  is a `TTY`, escape
    sequences will be used to highlight sections of the log.
    """

    def __init__(self, fd: int, total_steps: int = 0):
        super().__init__(fd, total_steps)
        self._module_start_time: Optional[float] = None

    def result(self, result: Union[StageResult, SourceResults], metadata: Optional[Dict] = None) -> None:
        if self._module_start_time is not None:
            duration = time.monotonic() - self._module_start_time
            self.out.write(f"\n⏱  Duration: {duration:.2f}s\n")

    def begin(self, pipeline):
        self.out.term(vt.bold, clear=True)
        self.out.write(f"Pipeline {pipeline.name}: {pipeline.id}")
        self.out.term(vt.reset)
        self.out.write("\n")
        self.out.write("Build\n  root: ")
        if pipeline.build:
            self.out.write(pipeline.build)
        else:
            self.out.write("<host>")
        if pipeline.runner:
            self.out.write(f"\n  runner: {pipeline.runner.name} ({pipeline.runner.exec})")
            if pipeline.run_in_vm:
                self.out.write(" in vm")
        source_epoch = pipeline.source_epoch
        if source_epoch is not None:
            timepoint = datetime.datetime.fromtimestamp(source_epoch).strftime('%c')
            self.out.write(f"\n  source-epoch: {timepoint} [{source_epoch}]")
        self.out.write("\n")

    def stage(self, stage):
        self.module(stage)

    def assembler(self, assembler):
        self.out.term(vt.bold, clear=True)
        self.out.write("Assembler ")
        self.out.term(vt.reset)

        self.module(assembler)

    def module(self, module):
        options = module.options or {}
        title = f"{module.name}: {module.id}"

        self.out.term(vt.bold, clear=True)
        self.out.write(title)
        self.out.term(vt.reset)
        self.out.write(" ")

        json.dump(options, self.out, indent=2)
        self.out.write("\n")

        self._module_start_time = time.monotonic()

    def log(self, message, origin: Optional[str] = None):
        self.out.write(message)


class JSONSeqMonitor(BaseMonitor):
    """Monitor that prints the log output of modules wrapped in json-seq objects with context and progress metadata"""

    def __init__(self, fd: int, total_steps: int):
        super().__init__(fd, total_steps)
        self._ctx_ids: Set[str] = set()
        self._progress = Progress("pipelines/sources", total_steps)
        self._context = Context(origin="org.osbuild")
        self._module_options: Dict[str, Any] = {}
        self._jsonseq_mu = Lock()
        self._module_start_time: Optional[float] = None

    def begin(self, pipeline: osbuild.Pipeline):
        self._context.set_pipeline(pipeline)
        if pipeline.stages:
            self._progress.sub_progress = Progress(f"pipeline: {pipeline.name}", len(pipeline.stages))
        self.log(f"Starting pipeline {pipeline.name}", origin="osbuild.monitor")

    # finish is for pipelines
    def finish(self, results: dict):
        self._progress.incr()
        self.log(f"Finished pipeline {results['name']}", origin="osbuild.monitor")

    def stage(self, stage: osbuild.Stage):
        self._module(stage)

    def assembler(self, assembler: osbuild.Stage):
        self._module(assembler)

    def _module(self, module: osbuild.Stage):
        self._context.set_stage(module)
        self._module_options = copy.deepcopy(module.options) if module.options else {}
        self.log(f"Starting module {module.name}", origin="osbuild.monitor")
        self._module_start_time = time.monotonic()

    def result(self, result: Union[StageResult, SourceResults], metadata: Optional[Dict] = None) -> None:
        """ Called when the module (stage or download) is finished

        This will stream a log entry that the stage finished and the result
        is available via the json-seq monitor as well. Note that while the
        stage output is part of the result it may be abbreviated. To get
        an entire buildlog the consumer needs to simply log the calls to
        "log()" which contain more detailed information as well.
        """
        # we may need to check pipeline ids here in the future
        if self._progress.sub_progress:
            self._progress.sub_progress.incr()

        # Limit the output in the json pipeline to a "reasonable"
        # length. We ran into an issue from a combination of a stage
        # that produce tons of output (~256 kb, see issue#1976) and
        # the consumer that used a golang scanner with a max default
        # buffer of 64kb before erroring.
        #
        # Consumers can collect the individual log lines on their own
        # if desired via the "log()" method.
        max_output_len = 31_000
        log_result = result
        # pylint: disable-next=unknown-option-value
        if isinstance(result, StageResult) and len(result.output) > max_output_len:  # type: ignore
            removed = len(result.output) - max_output_len
            log_result.output = f"[...{removed} bytes hidden...]\n{result.output[removed:]}"  # type: ignore
        elif isinstance(result, SourceResults) and len(json.dumps(result.as_dict())) > max_output_len:
            # Ignore the results of individual source items and just log
            # the overall result. The result generated by osbuild at the
            # end should have all the details in case of failures.
            log_result = SourceResults(result.source_type)
            log_result.success = result.success
            log_result.results = []

        # If the module has a timestamp, we can calculate the duration.
        duration = None
        if self._module_start_time is not None:
            duration = time.monotonic() - self._module_start_time

        self._jsonseq(log_entry(
            f"Finished module {result.name}",
            context=self._context.with_origin("osbuild.monitor"),
            progress=self._progress,
            options=self._module_options,
            duration=duration,
            # We should probably remove the "output" key from the result
            # as it is redundant, each output already generates a "log()"
            # message that is streamed to the client.
            result=log_result,
            metadata=metadata,
        ))

    def log(self, message, origin: Optional[str] = None):
        self._jsonseq(log_entry(
            message,
            context=self._context.with_origin(origin),
            progress=self._progress,
        ))

    def _jsonseq(self, entry: dict) -> None:
        with self._jsonseq_mu:
            # follow rfc7464 (application/json-seq)
            self.out.write("\x1e")
            json.dump(entry, self.out)
            self.out.write("\n")


def make(name: str, fd: int, total_steps: int) -> BaseMonitor:
    module = sys.modules[__name__]
    monitor = getattr(module, name, None)
    if not monitor:
        raise ValueError(f"Unknown monitor: {name}")
    if not issubclass(monitor, BaseMonitor):
        raise ValueError(f"Invalid monitor: {name}")
    return monitor(fd, total_steps)



#!/usr/bin/python3
# pylint: disable=redefined-outer-name

import contextlib
import copy
import os
from unittest.mock import MagicMock, patch

import pytest

STAGE_NAME = "org.osbuild.bfb"


FAKE_INPUTS = {
    "kernel": {
        "path": "/input/kernel/path",
        "data": {
            "files": {
                "kernel-file": {}
            }
        }
    },
    "initramfs": {
        "path": "/input/initramfs/path",
        "data": {
            "files": {
                "initramfs-file": {}
            }
        }
    }
}

FAKE_INPUTS_WITH_ROOTFS = {
    **FAKE_INPUTS,
    "rootfs": {
        "path": "/input/rootfs/path",
        "data": {
            "files": {
                "rootfs-file": {}
            }
        }
    }
}


@pytest.fixture
def mock_makeefi(stage_module):
    def fake_mkefidir(efidir, source_tree):  # pylint: disable=unused-argument
        os.makedirs(os.path.join(efidir, "BOOT"), exist_ok=True)

    def fake_mkefiboot(efidir, output_efiboot_img, loop_client):  # pylint: disable=unused-argument
        with open(output_efiboot_img, "wb") as f:
            f.write(b"fake_efiboot")

    with patch.object(stage_module.makeefi, "mkefidir",
                      side_effect=fake_mkefidir), \
            patch.object(stage_module.makeefi, "mkefiboot",
                         side_effect=fake_mkefiboot):
        yield


@pytest.fixture
def mock_loop_client():
    client = MagicMock()

    @contextlib.contextmanager
    def _fake_device(_path):
        yield "/dev/loop99"
    client.device = MagicMock(side_effect=_fake_device)
    return client


@pytest.mark.usefixtures("mock_makeefi")
@pytest.mark.parametrize("inputs,options,expected_cmd_parts", [
    # Basic test - kernel + initramfs only
    (
        FAKE_INPUTS,
        {"filename": "test.bfb"},
        [
            "/usr/bin/mlx-mkbfb",
            "--ramdisk",
            "--capsule",
            "--boot-path",
            "--boot-desc",
            "/lib/firmware/mellanox/boot/default.bfb",
        ]
    ),
    # Test with rootfs
    (
        FAKE_INPUTS_WITH_ROOTFS,
        {"filename": "test.bfb"},
        [
            "/usr/bin/mlx-mkbfb",
            "--ramdisk",
            "--capsule",
            "--boot-path",
            "--boot-desc",
            "/lib/firmware/mellanox/boot/default.bfb",
        ]
    ),
])
@patch("subprocess.run")
def test_bfb_command_generation(
        mock_run,
        tmp_path,
        stage_module,
        mock_loop_client,
        inputs,
        options,
        expected_cmd_parts):
    """Test that stage generates correct mlx-mkbfb command"""

    output_dir = str(tmp_path / "output")
    os.makedirs(output_dir)

    # Create fake input files so file I/O works
    test_inputs = copy.deepcopy(inputs)
    for name in test_inputs:
        inp = test_inputs[name]
        files = inp["data"]["files"]
        fname = list(files.keys())[0]
        inp["path"] = str(tmp_path / name)
        os.makedirs(inp["path"], exist_ok=True)
        with open(os.path.join(inp["path"], fname), "wb") as f:
            f.write(b"fake")

    stage_module.main(test_inputs, output_dir, options, mock_loop_client)

    actual_cmd = mock_run.call_args_list[-1][0][0]

    for part in expected_cmd_parts:
        assert part in actual_cmd, f"Expected {part!r} in command: {actual_cmd}"
    assert f"{output_dir}/{options['filename']}" in actual_cmd


@pytest.mark.usefixtures("mock_makeefi")
@patch("subprocess.run")
def test_bfb_rootfs_combination(mock_run, tmp_path, mock_loop_client, stage_module):
    """Test that initramfs and rootfs are combined when rootfs is provided"""

    options = {"filename": "test.bfb"}
    output_dir = str(tmp_path / "output")
    os.makedirs(output_dir)

    test_inputs = copy.deepcopy(FAKE_INPUTS_WITH_ROOTFS)
    for name in test_inputs:
        inp = test_inputs[name]
        fname = list(inp["data"]["files"].keys())[0]
        inp["path"] = str(tmp_path / name)
        os.makedirs(inp["path"], exist_ok=True)
        with open(os.path.join(inp["path"], fname), "wb") as f:
            f.write(b"fake_data")

    stage_module.main(test_inputs, output_dir, options, mock_loop_client)

    # mlx-mkbfb is the last call
    actual_cmd = mock_run.call_args_list[-1][0][0]
    assert "--ramdisk" in actual_cmd


@pytest.mark.usefixtures("mock_makeefi")
@patch("subprocess.run")
def test_bfb_no_rootfs(mock_run, tmp_path, mock_loop_client, stage_module):
    """Test that initramfs is used directly when rootfs is not provided"""

    options = {"filename": "test.bfb"}
    output_dir = str(tmp_path / "output")
    os.makedirs(output_dir)

    test_inputs = copy.deepcopy(FAKE_INPUTS)
    for name in test_inputs:
        inp = test_inputs[name]
        fname = list(inp["data"]["files"].keys())[0]
        inp["path"] = str(tmp_path / name)
        os.makedirs(inp["path"], exist_ok=True)
        with open(os.path.join(inp["path"], fname), "wb") as f:
            f.write(b"fake_data")

    stage_module.main(test_inputs, output_dir, options, mock_loop_client)

    actual_cmd = mock_run.call_args_list[-1][0][0]
    assert "--ramdisk" in actual_cmd


def test_parse_input(stage_module):
    """Test the parse_input helper function"""

    test_inputs = {
        "test": {
            "path": "/test/path",
            "data": {
                "files": {
                    "testfile": {}
                }
            }
        }
    }

    result = stage_module.parse_input(test_inputs, "test")
    assert result == "/test/path/testfile"



"""
Pipeline inputs

A pipeline input provides data in various forms to a `Stage`, like
files, OSTree commits or trees. The content can either be obtained
via a `Source` or have been built by a `Pipeline`. Thus an `Input`
is the bridge between various types of content that originate from
different types of sources.

The acceptable origin of the data is determined by the `Input`
itself. What types of input are allowed and required is determined
by the `Stage`.

To osbuild itself this is all transparent. The only data visible to
osbuild is the path. The input options are just passed to the
`Input` as is and the result is forwarded to the `Stage`.
"""

import abc
import hashlib
import json
import os
from pathlib import Path
from typing import Any, Dict, Optional, Tuple

from osbuild import host
from osbuild.util.types import PathLike

from .objectstore import StoreClient, StoreServer


class Input:
    """
    A single input with its corresponding options.
    """

    def __init__(self, name, info, origin: str, options: Dict):
        self.name = name
        self.info_name = info.name
        self.info_path = info.path
        self.origin = origin
        self.refs: Dict[str, Dict[str, Any]] = {}
        self.options = options or {}
        self.id = self.calc_id()

    def add_reference(self, ref, options: Optional[Dict] = None):
        self.refs[ref] = options or {}
        self.id = self.calc_id()

    def calc_id(self):
        # NB: The input `name` is not included here on purpose since it
        # is either prescribed by the stage itself and thus not actual
        # parameter or arbitrary and chosen by the manifest generator
        # and thus can be changed without affecting the contents
        m = hashlib.sha256()
        m.update(json.dumps(self.info_name, sort_keys=True).encode())
        m.update(json.dumps(self.origin, sort_keys=True).encode())
        m.update(json.dumps(self.refs, sort_keys=True).encode())
        m.update(json.dumps(self.options, sort_keys=True).encode())
        return m.hexdigest()

    def to_dict(self, libdir) -> Dict:
        return {
            "name": self.name,
            "info_name": self.info_name,
            "info_path": str(Path(self.info_path).relative_to(libdir)),
            "origin": self.origin,
            "refs": self.refs,
            "options": self.options,
        }

    @classmethod
    def from_dict(cls, data: Dict, libdir) -> "Input":
        class InfoStub:
            def __init__(self, name: str, path: str):
                self.name = name
                self.path = path

        info = InfoStub(data["info_name"], Path(libdir) / data["info_path"])
        inp = cls(
            name=data["name"],
            info=info,
            origin=data["origin"],
            options=data["options"]
        )
        inp.refs = data["refs"]
        inp.id = inp.calc_id()
        return inp


class InputManager:
    def __init__(self, mgr: host.ServiceManager, storeapi: StoreServer, root: PathLike) -> None:
        self.service_manager = mgr
        self.storeapi = storeapi
        self.root = root
        self.inputs: Dict[str, Input] = {}

    def map(self, ip: Input) -> Tuple[str, Dict]:

        target = os.path.join(self.root, ip.name)
        os.makedirs(target)

        args = {
            # mandatory bits
            "origin": ip.origin,
            "refs": ip.refs,

            "target": target,

            # global options
            "options": ip.options,

            # API endpoints
            "api": {
                "store": self.storeapi.socket_address
            }
        }

        client = self.service_manager.start(f"input/{ip.name}", ip.info_path)
        reply = client.call("map", args)

        path = reply["path"]

        if not path.startswith(self.root):
            raise RuntimeError(f"returned {path} has wrong prefix")

        reply["path"] = os.path.relpath(path, self.root)

        self.inputs[ip.name] = reply

        return reply


class InputService(host.Service):
    """Input host service"""

    @abc.abstractmethod
    def map(self, store, origin, refs, target, options):
        pass

    def unmap(self):
        pass

    def stop(self):
        self.unmap()

    def dispatch(self, method: str, args, fds):
        if method == "map":
            store = StoreClient(connect_to=args["api"]["store"])
            r = self.map(store,
                         args["origin"],
                         args["refs"],
                         args["target"],
                         args["options"])
            return r, None

        raise host.ProtocolError("Unknown method")



"""Module for creating SPDX spec v2 Software Bill of Materials (SBOM) files."""

from .model import (
    Checksum,
    ChecksumAlgorithm,
    CreationInfo,
    Creator,
    CreatorType,
    Document,
    ExternalPackageRef,
    ExternalPackageRefCategory,
    ExtractedLicensingInfo,
    NoAssertionValue,
    NoneValue,
    Package,
    Relationship,
    RelationshipType,
)

__all__ = [
    "Checksum",
    "ChecksumAlgorithm",
    "CreationInfo",
    "Creator",
    "CreatorType",
    "Document",
    "ExternalPackageRef",
    "ExtractedLicensingInfo",
    "ExternalPackageRefCategory",
    "NoAssertionValue",
    "NoneValue",
    "Package",
    "Relationship",
    "RelationshipType"
]



#!/usr/bin/python3

STAGE_NAME = "org.osbuild.dracut.conf"


def test_dracut_conf_smoke(tmp_path, stage_module):
    options = {
        "filename": "qemu.conf",
        "config": {
            "add_drivers": ["virtio_blk"],
        }
    }

    tree = tmp_path / "tree"
    stage_module.main(tree, options)
    expected_dracut_conf = tree / "usr/lib/dracut/dracut.conf.d/qemu.conf"
    assert 'add_drivers+=" virtio_blk "\n' == expected_dracut_conf.read_text()



#!/usr/bin/python3

import tempfile
from contextlib import contextmanager
from unittest.mock import Mock, call, patch

import pytest

STAGE_NAME = "org.osbuild.bootc.install-to-filesystem"


@pytest.fixture(name="mocked_named_tmp")
def mocked_named_tmp_fixture():
    fake_named_tmp = Mock()
    fake_named_tmp.name = "/tmp/fake-named-tmpfile-name"
    with patch("tempfile.NamedTemporaryFile", return_value=fake_named_tmp):
        yield


@pytest.fixture(name="mocked_temp_dir")
def mocked_temp_dir_fixture(tmp_path):
    @contextmanager
    def _tmp_dir():
        yield tmp_path
    with patch("tempfile.TemporaryDirectory", side_effect=_tmp_dir):
        yield tmp_path


FAKE_INPUTS = {
    "images": {
        "path": "/input/images/path",
        "data": {
            "archives": {
                "filename": {
                    "format": "oci-archive",
                    "name": "some-img-name",
                },
            },
        },
    }
}


@pytest.mark.parametrize("options,expected_args", [
    ({}, []),
    # root-ssh
    ({"root-ssh-authorized-keys": []}, []),
    ({"root-ssh-authorized-keys": ["ssh-key"]}, ["--root-ssh-authorized-keys", "/tmp/fake-named-tmpfile-name"]),
    ({"root-ssh-authorized-keys": ["key1", "key2"]}, ["--root-ssh-authorized-keys", "/tmp/fake-named-tmpfile-name"]),
    # kernel args
    ({"kernel-args": []}, []),
    ({"kernel-args": ["console=ttyS0"]}, ["--karg", "console=ttyS0"]),
    ({"kernel-args": ["arg1", "arg2"]}, ["--karg", "arg1", "--karg", "arg2"]),
    # stateroot
    ({"stateroot": ""}, []),
    ({"stateroot": "default1"}, ["--stateroot", "default1"]),
    # root-mount-spec
    ({"root-mount-spec": ""}, ["--root-mount-spec="]),
    ({"root-mount-spec": "subvol@root"}, ["--root-mount-spec", "subvol@root"]),
    # boot-mount-spec
    ({"boot-mount-spec": ""}, ["--boot-mount-spec="]),
    ({"boot-mount-spec": "/dev/sda1"}, ["--boot-mount-spec", "/dev/sda1"]),
    # bootupd-skip-boot-uuid
    ({"bootupd-skip-boot-uuid": False}, []),
    ({"bootupd-skip-boot-uuid": True}, ["--bootupd-skip-boot-uuid"]),
    # bootloader
    ({"bootloader": "none"}, ["--bootloader", "none"]),
    ({"bootloader": "systemd"}, ["--bootloader", "systemd"]),
    ({"bootloader": "grub"}, ["--bootloader", "grub"]),
    # composefs
    ({"composefs": False}, []),
    ({"composefs": True}, ["--composefs-backend"]),
    # all
    ({"root-ssh-authorized-keys": ["key1", "key2"],
      "kernel-args": ["arg1", "arg2"],
      "target-imgref": "quay.io/img/ref",
      "stateroot": "/some/stateroot",
      "root-mount-spec": "root-mount-spec",
      "boot-mount-spec": "boot-mount-spec",
      "bootupd-skip-boot-uuid": True,
      },
     ["--root-ssh-authorized-keys", "/tmp/fake-named-tmpfile-name",
      "--karg", "arg1", "--karg", "arg2",
      "--target-imgref", "quay.io/img/ref",
      "--stateroot", "/some/stateroot",
      "--root-mount-spec", "root-mount-spec",
      "--boot-mount-spec", "boot-mount-spec",
      "--bootupd-skip-boot-uuid",
      ],
     ),
])
@patch("subprocess.run")
@patch("subprocess.check_call")
def test_bootc_install_to_fs(mock_check_call, mock_run, mocked_named_tmp, mocked_temp_dir, stage_module, options, expected_args):  # pylint: disable=unused-argument
    inputs = {
        "images": {
            "path": "/input/images/path",
            "data": {
                "archives": {
                    "filename": {
                        "format": "oci-archive",
                        "name": "some-img-name",
                    },
                },
            },
        },
    }
    paths = {
        "mounts": "/path/to/mounts",
    }

    stage_module.main(options, inputs, paths)

    assert len(mock_run.call_args_list) == 1
    args, kwargs = mock_run.call_args_list[0]
    assert args == (
        ["bootc", "install", "to-filesystem",
         "--source-imgref", f"oci-archive:{mocked_temp_dir}/image",
         "--skip-fetch-check", "--generic-image",
         ] + expected_args + ["/path/to/mounts"],
    )
    assert kwargs["check"] is True
    assert kwargs["env"]["BOOTC_SKIP_SELINUX_HOST_CHECK"] == "true"

    assert mock_check_call.call_count == 4
    mock_check_call.assert_has_calls([
        call(['mount', '-t', 'devtmpfs', 'devtmpfs', '/dev/']),
        call(['mount', '-t', 'tmpfs', 'tmpfs', '/dev/shm']),
        call(['umount', '/dev/shm']),
        call(['umount', '/dev/']),
    ])


@patch("subprocess.run")
@patch("subprocess.check_call")
def test_bootc_install_to_fs_write_root_ssh_keys(_mock_check_call, mock_run, stage_module):  # pylint: disable=unused-argument
    paths = {
        "mounts": "/path/to/mounts",
    }
    options = {
        "root-ssh-authorized-keys": ["key1", "key2"],
    }

    named_tmp = tempfile.NamedTemporaryFile(delete=False)
    with patch("tempfile.NamedTemporaryFile", return_value=named_tmp):
        stage_module.main(options, FAKE_INPUTS, paths)
        with open(named_tmp.name, encoding="utf8") as fp:
            assert "key1\nkey2\n" == fp.read()



"""
Mount Handling for pipeline stages

Allows stages to access file systems provided by devices.
This makes mount handling transparent to the stages, i.e.
the individual stages do not need any code for different
file system types and the underlying devices.
"""

import abc
import hashlib
import json
import os
import subprocess
from pathlib import Path
from typing import Dict, List, Union

from osbuild import host
from osbuild.devices import DeviceManager
from osbuild.mixins import MixinImmutableID


class Mount(MixinImmutableID):
    """
    A single mount with its corresponding options
    """

    def __init__(self, name, info, device, partition, target, options: Dict):
        self.name = name
        self.info_name = info.name
        self.info_path = info.path
        self.device = device
        self.partition = partition
        self.target = target
        self.options = options
        self.id = self.calc_id()

    def calc_id(self):
        m = hashlib.sha256()
        m.update(json.dumps(self.info_name, sort_keys=True).encode())
        if self.device:
            m.update(json.dumps(self.device.id, sort_keys=True).encode())
        if self.partition:
            m.update(json.dumps(self.partition, sort_keys=True).encode())
        if self.target:
            m.update(json.dumps(self.target, sort_keys=True).encode())
        m.update(json.dumps(self.options, sort_keys=True).encode())
        return m.hexdigest()

    def to_dict(self, libdir) -> Dict:
        return {
            "name": self.name,
            "info_name": self.info_name,
            "info_path": str(Path(self.info_path).relative_to(libdir)),
            "device_name": self.device.name if self.device else None,
            "partition": self.partition,
            "target": self.target,
            "options": self.options,
        }

    @classmethod
    def from_dict(cls, data: Dict, libdir, device=None) -> "Mount":
        class InfoStub:
            def __init__(self, name: str, path: str):
                self.name = name
                self.path = path

        info = InfoStub(data["info_name"], Path(libdir) / data["info_path"])
        mnt = cls(
            name=data["name"],
            info=info,
            device=device,
            partition=data["partition"],
            target=data["target"],
            options=data["options"]
        )
        return mnt


class MountManager:
    """Manager for Mounts

    Uses a `host.ServiceManager` to activate `Mount` instances.
    Takes a `DeviceManager` to access devices and a directory
    called `root`, which is the root of all the specified mount
    points.
    """

    def __init__(self, devices: DeviceManager, root: str) -> None:
        self.devices = devices
        self.root = root
        self.mounts: Dict[str, Union[str, Dict[str, Mount]]] = {}

    def mount(self, mount: Mount) -> Dict:

        # Get the absolute path to the source device inside the
        # temporary filesystem (i.e. /run/osbuild/osbuild-dev-xyz/loop0)
        # and also the relative path to the source device inside
        # that filesystem (i.e. loop0). If the device also exists on the
        # host in `/dev` (like /dev/loop0), we'll use that path for the
        # mount because some tools (like grub2-install) consult mountinfo
        # to try to canonicalize paths for mounts and inside the bwrap env
        # the device will be under `/dev`. https://github.com/osbuild/osbuild/issues/1492
        source = self.devices.device_abspath(mount.device)
        relpath = self.devices.device_relpath(mount.device)
        if relpath and os.path.exists(os.path.join('/dev', relpath)):
            source = os.path.join('/dev', relpath)

        # If the user specified a partition then the filesystem to
        # mount is actually on a partition of the disk.
        if source and mount.partition:
            source = f"{source}p{mount.partition}"

        root = os.fspath(self.root)

        args = {
            "source": source,
            "target": mount.target,

            "root": root,
            "tree": os.fspath(self.devices.tree),

            "options": mount.options,
        }

        mgr = self.devices.service_manager

        client = mgr.start(f"mount/{mount.name}", mount.info_path)
        path = client.call("mount", args)

        if not path:
            res: Dict[str, Mount] = {}
            self.mounts[mount.name] = res
            return res

        if not path.startswith(root):
            raise RuntimeError(f"returned path '{path}' has wrong prefix")

        path = os.path.relpath(path, root)

        self.mounts[mount.name] = path

        return {"path": path}


class MountService(host.Service):
    """Mount host service"""

    @abc.abstractmethod
    def mount(self, args: Dict):
        """Mount a device"""

    @abc.abstractmethod
    def umount(self):
        """Unmount all mounted resources"""

    def stop(self):
        self.umount()

    def dispatch(self, method: str, args, _fds):
        if method == "mount":
            r = self.mount(args)
            return r, None

        raise host.ProtocolError("Unknown method")


class FileSystemMountService(MountService):
    """Specialized mount host service for file system mounts"""

    def __init__(self, args):
        super().__init__(args)

        self.mountpoint = None
        self.check = False

    # pylint: disable=no-self-use
    @abc.abstractmethod
    def translate_options(self, options: Dict) -> List:
        opts = []
        if options.get("readonly"):
            opts.append("ro")
        if options.get("norecovery"):
            opts.append("norecovery")
        if "uid" in options:
            opts.append(f"uid={options['uid']}")
        if "gid" in options:
            opts.append(f"gid={options['gid']}")
        if "umask" in options:
            opts.append(f"umask={options['umask']}")
        if "shortname" in options:
            opts.append(f"shortname={options['shortname']}")
        if "subvol" in options:
            opts.append(f"subvol={options['subvol']}")
        if "compress" in options:
            opts.append(f"compress={options['compress']}")
        if opts:
            return ["-o", ",".join(opts)]
        return []

    def mount(self, args: Dict):

        source = args["source"]
        target = args["target"]
        root = args["root"]
        options = args["options"]

        mountpoint = os.path.join(root, target.lstrip("/"))

        options = self.translate_options(options)

        os.makedirs(mountpoint, exist_ok=True)
        self.mountpoint = mountpoint

        print(f"mounting {source} -> {mountpoint}")

        try:
            subprocess.run(
                ["mount"] +
                options + [
                    "--source", source,
                    "--target", mountpoint
                ],
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE,
                check=True)
        except subprocess.CalledProcessError as e:
            code = e.returncode
            msg = e.stdout.strip()
            raise RuntimeError(f"{msg} (code: {code})") from e

        self.check = True
        return mountpoint

    def umount(self):
        if not self.mountpoint:
            return

        # It's possible this mountpoint has already been unmounted
        # if a umount -R was run by another process, as is done in
        # mounts/org.osbuild.ostree.deployment.
        if not os.path.ismount(self.mountpoint):
            print(f"already unmounted: {self.mountpoint}")
            return

        self.sync()

        # We ignore errors here on purpose
        subprocess.run(["umount", "-v", self.mountpoint],
                       check=self.check)
        self.mountpoint = None

    def sync(self):
        subprocess.run(["sync", "-f", self.mountpoint],
                       check=self.check)



#!/usr/bin/python3

import os
import os.path

STAGE_NAME = "org.osbuild.systemd"


def test_systemd_masked_generators(tmp_path, stage_module):
    options = {"masked_generators": ["fake-generator"]}

    stage_module.main(tmp_path, options)

    masked_generator_path = os.path.join(tmp_path, "etc/systemd/system-generators/fake-generator")
    assert os.path.lexists(masked_generator_path)
    assert os.readlink(masked_generator_path) == "/dev/null"



import os
from unittest import mock

import pytest

from osbuild import testutil

STAGE_NAME = "org.osbuild.rpm"


@pytest.mark.parametrize(
    "test_data,expected_err",
    [
        # bad
        ({"kernel_install_env": {"boot_root": "../rel"}}, "'../rel' does not match "),
        ({"kernel_install_env": {"conf_root": "../rel"}}, "'../rel' does not match "),
        ({"kernel_install_env": {"machine_id": "not-a-valid-id"}}, "'not-a-valid-id' does not match "),
        ({"kernel_install_env": {"machine_id": "ABCD0123abcd0123abcd0123abcd0123"}}, "does not match "),
        ({"nodeps": "yes"}, "is not of type 'boolean'"),
        ({"generic_env": "not-an-object"}, "is not of type 'object'"),
        ({"generic_env": {"FOO": 123}}, "is not of type 'string'"),
        # good
        ({}, ""),
        ({"kernel_install_env": {"boot_root": "/boot"}}, ""),
        ({"kernel_install_env": {"conf_root": "/etc/kernel"}}, ""),
        ({"kernel_install_env": {"plugins": "/usr/lib/kernel/install.d/50-depmod.install"}}, ""),
        ({"kernel_install_env": {"plugins": ":"}}, ""),
        ({"kernel_install_env": {"machine_id": "abcd0123abcd0123abcd0123abcd0123"}}, ""),
        ({"nodeps": True}, ""),
        ({"nodeps": False}, ""),
        ({"generic_env": {}}, ""),
        ({"generic_env": {"FOO": "bar"}}, ""),
        ({"generic_env": {"FOO": "bar", "BAZ": "qux"}}, ""),
    ],
)
def test_schema_validation(stage_schema, test_data, expected_err):
    test_input = {"type": STAGE_NAME, "options": {}}
    test_input["options"].update(test_data)
    res = stage_schema.validate(test_input)

    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


@pytest.mark.parametrize(
    "generic_env,kernel_install_env,expected_err",
    [
        ({"BOOT_ROOT": "/other"}, {"boot_root": "/boot"}, "BOOT_ROOT"),
        (
            {"MACHINE_ID": "abcd0123abcd0123abcd0123abcd0123"},
            {"machine_id": "abcd0123abcd0123abcd0123abcd0123"},
            "MACHINE_ID",
        ),
        (
            {"BOOT_ROOT": "/a", "MACHINE_ID": "abcd0123abcd0123abcd0123abcd0123"},
            {"boot_root": "/b", "machine_id": "abcd0123abcd0123abcd0123abcd0123"},
            "BOOT_ROOT.*MACHINE_ID",
        ),
    ],
)
@mock.patch("subprocess.run")
def test_env_conflict_detection(tmp_path, stage_module, generic_env, kernel_install_env, expected_err):
    tree = str(tmp_path / "tree")
    os.makedirs(tree)
    inputs = {"packages": {"path": str(tmp_path), "data": {"files": {}}}}
    options = {
        "generic_env": generic_env,
        "kernel_install_env": kernel_install_env,
    }
    with pytest.raises(ValueError, match=expected_err):
        stage_module.main(tree, inputs, options)


@pytest.mark.parametrize(
    "ignore_failures",
    [
        (False),
        (True),
    ],
)
@mock.patch("subprocess.run")
def test_import_gpg_keys(mock_run, tmp_path, stage_module, ignore_failures):
    tree = tmp_path / "tree"
    tree.mkdir()
    stage_module.import_gpg_keys(tree, ["some-key"], "my-binary", [], ignore_failures)
    mock_run.assert_called_once()
    assert mock_run.call_args[0][0][0] == "my-binary"
    assert mock_run.call_args[1] == {"check": not ignore_failures}


@pytest.mark.parametrize(
    "rpm_output,expected_packages",
    [
        # all optional fields present
        (
            "{\n"
            '"name": "bash",\n'
            '"version": "5.2.26",\n'
            '"release": "3.fc40",\n'
            '"epoch": "0",\n'
            '"arch": "x86_64",\n'
            '"sigmd5": "abc123",\n'
            '"sha1header": "da39a3ee5e6b4b0d3255bfef95601890afd80709",\n'
            '"sha256header": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",\n'
            '"sha3_256header": "a7ffc6f8bf1ed76651c14756a061d662f580ff4de43b49fa82d80a4b80f8434a",\n'
            '"sigpgp": "deadbeef",\n'
            '"siggpg": null\n'
            "},\n",
            [
                {
                    "name": "bash",
                    "version": "5.2.26",
                    "release": "3.fc40",
                    "epoch": "0",
                    "arch": "x86_64",
                    "sigmd5": "abc123",
                    "sha1header": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
                    "sha256header": "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
                    "sha3_256header": "a7ffc6f8bf1ed76651c14756a061d662f580ff4de43b49fa82d80a4b80f8434a",
                    "sigpgp": "deadbeef",
                    "siggpg": None,
                }
            ],
        ),
        # all optional fields null (including new header digests)
        (
            "{\n"
            '"name": "coreutils",\n'
            '"version": "9.4",\n'
            '"release": "1.fc40",\n'
            '"epoch": null,\n'
            '"arch": null,\n'
            '"sigmd5": null,\n'
            '"sha1header": null,\n'
            '"sha256header": null,\n'
            '"sha3_256header": null,\n'
            '"sigpgp": null,\n'
            '"siggpg": null\n'
            "},\n",
            [
                {
                    "name": "coreutils",
                    "version": "9.4",
                    "release": "1.fc40",
                    "epoch": None,
                    "arch": None,
                    "sigmd5": None,
                    "sha1header": None,
                    "sha256header": None,
                    "sha3_256header": None,
                    "sigpgp": None,
                    "siggpg": None,
                }
            ],
        ),
        # multiple packages, sorted by name in output
        (
            "{\n"
            '"name": "zsh",\n'
            '"version": "5.9",\n'
            '"release": "1.fc40",\n'
            '"epoch": null,\n'
            '"arch": "x86_64",\n'
            '"sigmd5": null,\n'
            '"sha1header": null,\n'
            '"sha256header": "abc",\n'
            '"sha3_256header": null,\n'
            '"sigpgp": null,\n'
            '"siggpg": null\n'
            "},\n"
            "{\n"
            '"name": "bash",\n'
            '"version": "5.2",\n'
            '"release": "1.fc40",\n'
            '"epoch": null,\n'
            '"arch": "x86_64",\n'
            '"sigmd5": null,\n'
            '"sha1header": "def",\n'
            '"sha256header": null,\n'
            '"sha3_256header": null,\n'
            '"sigpgp": null,\n'
            '"siggpg": null\n'
            "},\n",
            [
                {
                    "name": "bash",
                    "version": "5.2",
                    "release": "1.fc40",
                    "epoch": None,
                    "arch": "x86_64",
                    "sigmd5": None,
                    "sha1header": "def",
                    "sha256header": None,
                    "sha3_256header": None,
                    "sigpgp": None,
                    "siggpg": None,
                },
                {
                    "name": "zsh",
                    "version": "5.9",
                    "release": "1.fc40",
                    "epoch": None,
                    "arch": "x86_64",
                    "sigmd5": None,
                    "sha1header": None,
                    "sha256header": "abc",
                    "sha3_256header": None,
                    "sigpgp": None,
                    "siggpg": None,
                },
            ],
        ),
    ],
)
@mock.patch("subprocess.run")
def test_generate_package_metadata(mock_run, tmp_path, stage_module, rpm_output, expected_packages):
    mock_run.return_value = mock.Mock(stdout=rpm_output)
    tree = str(tmp_path / "tree")
    result = stage_module.generate_package_metadata(tree, [])
    assert result["packages"] == expected_packages
    cmd = mock_run.call_args[0][0]
    assert cmd[0] == "rpm"
    assert "--root" in cmd
    assert "-qa" in cmd



"""File System Cache

This module implements a data cache that uses the file system to store data
as well as protect parallel access. It implements automatic cache management
and allows purging the cache during runtime, pruning old entries and keeping
the cache under a given limit.
"""

# pylint: disable=too-many-lines

import contextlib
import ctypes
import errno
import json
import os
import subprocess
import uuid
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

from osbuild.util import ctx, linux, rmrf

__all__ = [
    "FsCache",
    "FsCacheInfo",
]


MaximumSizeType = Optional[Union[int, str]]


class FsCacheInfo(NamedTuple):
    """File System Cache Information

    This type represents static cache information. It is an immutable named
    tuple and used to query or set the configuration of a cache.

    creation_boot_id - Hashed linux boot-id at the time of cache-creation
    maximum_size - Maximum cache size in bytes, or "unlimited"
    version - version of the cache data structures
    """

    creation_boot_id: Optional[str] = None
    maximum_size: MaximumSizeType = None
    version: Optional[int] = None

    @classmethod
    def from_json(cls, data: Any) -> "FsCacheInfo":
        """Create tuple from parsed JSON

        This takes a parsed JSON value and converts it into a tuple with the
        same information. Unknown fields in the input are ignored. The input
        is usually taken from `json.load()` and similar.
        """

        if not isinstance(data, dict):
            return cls()

        creation_boot_id = None
        maximum_size: MaximumSizeType = None
        version = None

        # parse "creation-boot-id"
        _creation_boot_id = data.get("creation-boot-id")
        if isinstance(_creation_boot_id, str) and len(_creation_boot_id) == 32:
            creation_boot_id = _creation_boot_id

        # parse "maximum-size"
        _maximum_size = data.get("maximum-size")
        if isinstance(_maximum_size, int):
            maximum_size = _maximum_size
        elif isinstance(_maximum_size, str) and _maximum_size == "unlimited":
            maximum_size = "unlimited"

        # parse "version"
        _version = data.get("version")
        if isinstance(_version, int):
            version = _version

        # create immutable tuple
        return cls(
            creation_boot_id,
            maximum_size,
            version,
        )

    def to_json(self) -> Dict[str, Any]:
        """Convert tuple into parsed JSON

        Return a parsed JSON value that represents the same values as this
        tuple does. Unset values are skipped. The returned value can be
        converted into formatted JSON via `json.dump()` and similar.
        """

        data: Dict[str, Any] = {}
        if self.creation_boot_id is not None:
            data["creation-boot-id"] = self.creation_boot_id
        if self.maximum_size is not None:
            data["maximum-size"] = self.maximum_size
        if self.version is not None:
            data["version"] = self.version
        return data


class FsCacheObjectInfo(NamedTuple):
    """ File System Cache object information

    This type represents information about a single cache object. The
    last_used information is only guaranteed to be valid while the cache
    is locked.
    """
    name: str
    last_used: float


class FsCache(contextlib.AbstractContextManager, os.PathLike):
    """File System Cache

    This file system cache context represents an on-disk cache. That is, it
    allows storing information on the file system, and retrieving it from other
    contexts.

    A single cache directory can be shared between many processes at the same
    time. The cache protects access to the cached data. The cache must not be
    shared over non-coherent network storage, but is designed for system-local
    linux file-systems.

    The file-system layout is as follows:

        [cache]/
        ├── cache.info
        ├── cache.lock
        ├── cache.size
        ├── objects/
        │   ├── [id0]
        │   ├── [id1]/
        │   │   ├── data/
        │   │   │   └── ...
        │   │   ├── object.info
        │   │   └── object.lock
        │   └── ...
        └── stage/
            ├── uuid-[uuid0]
            ├── uuid-[uuid1]/
            │   ├── data/
            │   │   └── ...
            │   ├── object.info
            │   └── object.lock
            └── ...

    The central data store is in the `objects` subdirectory. Every cache entry
    has a separate subdirectory there. To guard access, a read-lock on
    `object.lock` is required for all readers, a write-lock is required for all
    writers. Static information about the object is available in the
    `object.info` file.

    As an optimization, entries in the object store consisting of a single
    file can be stored directly underneath `objects` without a separate
    subdirectory hierarchy. Their guarding lock is directly taken on this file
    and no metadata is available, other than the file information itself. This
    is used extensively by the cache management to prepare objects for atomic
    replacements. Due to lack of metadata, they are volatile and can be
    deleted as soon as they are unlocked.

    Generally, access to the cache is non-blocking. That is, if a read-lock
    cannot be acquired, an entry is considered non-existant. Thus, unless
    treated as a `write-once` cache, cache efficiency will decrease when taking
    write-locks.

    The `data/` directory contains the content of a cache entry. Its content
    is solely defined by the creator of the entry and the cache makes no
    assumptions about its layout. Note that the `data/` directory itself can be
    modified (e.g., permission-changes) if an unnamed top-level directory is
    desired (e.g., to store a directory tree).

    Additionally to the `objects/` directory, a similar `stage/` directory is
    provided. This directory is `write-only` and used to prepare entries for
    the object store before committing them. The staging area is optional. It
    is completely safe to do the same directly in the object store. However,
    the separation allows putting the staging area on a different file-system
    (e.g., symlinking to a tmpfs), and thus improving performance for larger
    operations. Otherwise, the staging area follows the same rules as the
    object store, except that only writers are expected. Hence, staging entries
    always use a unique UUID as name. To commit a staging entry, a user is
    expected to create an entry in the object store and copy/move the `data/`
    directory over.

    To guard against parallel accesses, a set of locks is utilized. Generally,
    a `*.lock`-file locks the directory it is in, while a lock on any other
    file just locks that file (unfortunately, we cannot acquire write-locks on
    directories directly, since it would require opening them for writing,
    which is not possible on linux). `cache.lock` can be used to guard the
    entire cache. A write-lock will keep any other parallel operation out,
    while a read-lock merely acquires cache access (you are still allowed to
    modify the cache, but need fine-grained locking). Hence, a write-lock on the
    global `cache.lock` file is only required for operations that cannot use
    fine-grained locking. The latter requires individual locking for each file
    or each object store entry you modify. In all those cases you must ensure
    for parallel modifications, since lock acquisition on file-systems can only
    be done after opening a file.
    """

    class MissError(Exception):
        """Cache Miss Exception

        This error is raised when a cache entry is not found. Due to the
        shared nature of the cache, a caller must be aware that any entry can
        be created or deleted by other concurrent operations, at any point in
        time. Hence, a cache miss only reflects the state of the cache at a
        particular time under a particular lock.
        """

    # static parameters
    _dirname_data = "data"
    _dirname_objects = "objects"
    _dirname_stage = "stage"
    _filename_cache_info = "cache.info"
    _filename_cache_lock = "cache.lock"
    _filename_cache_size = "cache.size"
    _filename_cache_tag = "CACHEDIR.TAG"
    _filename_object_info = "object.info"
    _filename_object_lock = "object.lock"
    _version_current = 1
    _version_minimum = 1

    # constant properties
    _appid: str
    _tracers: Dict[str, Any]
    _path_cache: Any

    # context-manager properties
    _active: bool
    _bootid: Optional[str]
    _lock: Optional[int]
    _info: FsCacheInfo
    _info_maximum_size: int

    def __init__(self, appid: str, path_cache: Any):
        """Create File System Cache

        This creates a new file-system cache. It does not create the cache, nor
        access any of its content. You must enter its context-manager to prepare
        the cache for access. Any access outside of a context-manager will raise
        an assertion error, unless explicitly stated otherwise.

        Parameters:
        -----------
        appid
            The application-ID of the caller. This can be any random string. It
            is used to initialize the application-specific boot-ID used to tag
            caches and detect whether an entry was created during the same boot.
        path_cache
            The path to the cache directory. The directory (and the path to it)
            is created if it does not exist.
        """

        self._appid = appid
        self._tracers = {}
        self._path_cache = os.fspath(path_cache)

        self._active = False
        self._bootid = None
        self._lock = None
        self._info = FsCacheInfo()
        self._info_maximum_size = 0

    def _trace(self, trace: str):
        """Trace execution

        Execute registered trace-hooks for the given trace string. This allows
        tests to register callbacks that are executed at runtime at a specific
        location in the code. During normal operation, no such hooks should be
        used.

        The trace-hooks are used to trigger race-conditions during tests and
        verify they are handled gracefully.

        Parameters:
        -----------
        trace
            The trace-hook to run.
        """

        if trace in self._tracers:
            self._tracers[trace]()

    @staticmethod
    def _calculate_space(path_target: str) -> int:
        """Calculate total space of a directory tree

        Calculate the total amount of storage required for a directory tree in
        bytes. This does not account for metadata, but only for stored file
        content.

        Note that this may differ from the sum of the file sizes as it
        takes sparse files into account.

        Parameters:
        -----------
        path_target
            File-system path to the directory to operate on.
        """

        return os.lstat(path_target).st_blocks * 512 + sum(
            os.lstat(
                os.path.join(path, f)
            ).st_blocks * 512 for path, dirs, files in os.walk(
                path_target
            ) for f in files + dirs
        )

    def __fspath__(self) -> Any:
        """Return cache path

        Return the path to this cache as provided to the constructor of the
        cache. No conversions are applied, so the path is absolute if the
        path as provided by the caller was absolute, and vice-versa.

        This is part of the `os.PathLike` interface. See its documentation.
        """

        return self._path_cache

    def _path(self, *rpaths):
        """Return absolute path into cache location

        Take the relative path from the caller and turn it into an absolute
        path. Since most operations take a relative path from the cache root
        to a cache location, this function can be used to make those paths
        absolute.

        Parameters:
        -----------
        rpaths
            Relative paths from cache root to the desired cache location.
        """

        return os.path.join(self, *rpaths)

    @contextlib.contextmanager
    def _atomic_open(
        self,
        rpath: str,
        *,
        wait: bool,
        write: bool,
        closefd: bool = True,
        oflags: int = 0,
    ):
        """Atomically open and lock file

        Open the cache-file at the specified relative path and acquire a
        lock on it. Yield the file-descriptor to the caller. Once control
        returns, all locks are released (if not already done so by the
        caller) and the file-descriptor is closed.

        Note that this operation involves a retry-loop in case the file is
        replaced or moved before the lock is acquired.

        Parameters:
        -----------
        rpath
            Relative path from the cache-root to the file to open.
        wait
            Whether to wait for locks to be acquired.
        write
            If false, the file is opened for reading and a read lock is
            acquired. If true, it is opened for read and write and a write
            lock is acquired.
        closefd
            If false, retain file-descriptor (and lock) on success.
        oflags
            Additional open-flags to pass to `os.open()`.
        """

        fd = None
        path = self._path(rpath)

        try:
            while True:
                # Open the file and acquire a lock. Make sure not to modify the
                # file in any way, ever. If non-blocking operation was requested
                # the lock call will raise `EAGAIN` if contended.
                flags = os.O_RDONLY | os.O_CLOEXEC | oflags
                lock = linux.fcntl.F_RDLCK
                if write:
                    flags = flags | os.O_RDWR
                    lock = linux.fcntl.F_WRLCK
                self._trace("_atomic_open:open")
                fd = os.open(path, flags, 0o644)
                self._trace("_atomic_open:lock")
                linux.fcntl_flock(fd, lock, wait=wait)

                # The file might have been replaced between opening it and
                # acquiring the lock. Hence, run `stat(2)` on the path again
                # and compare it to `fstat(2)` of the open file. If they differ
                # simply retry.
                # On NFS, the lock-acquisition has invalidated the caches, hence
                # the metadata is refetched. On linux, the first query will
                # succeed and reflect the drop in link-count. Every further
                # query will yield `ESTALE`. Yet, we cannot rely on being the
                # first to query, so proceed carefully.
                # On non-NFS, information is coherent and we can simply proceed
                # comparing the DEV+INO information to see whether the file was
                # replaced.

                retry = False

                try:
                    st_fd = os.stat(fd)
                except OSError as e:
                    if e.errno != errno.ESTALE:
                        raise
                    retry = True

                try:
                    st_path = os.stat(path)
                except OSError as e:
                    if e.errno not in [errno.ENOENT, errno.ESTALE]:
                        raise
                    retry = True

                if retry or st_fd.st_dev != st_path.st_dev or st_fd.st_ino != st_path.st_ino:
                    linux.fcntl_flock(fd, linux.fcntl.F_UNLCK)
                    os.close(fd)
                    fd = None
                    continue

                # Yield control to the caller to make use of the FD. If the FD
                # is to be retained, clear it before returning to the cleanup
                # handlers.
                yield fd

                if not closefd:
                    fd = None

                return
        finally:
            if fd is not None:
                linux.fcntl_flock(fd, linux.fcntl.F_UNLCK)
                os.close(fd)

    @contextlib.contextmanager
    def _atomic_file(
        self,
        rpath: str,
        rpath_store: str,
        closefd: bool = True,
        ignore_exist: bool = False,
        replace: bool = False,
    ):
        """Create and link temporary file

        Create a new temporary file and yield control to the caller to fill in
        data and metadata. Once control is returned, the file is linked at the
        specified location. If an exception is raised, the temporary file is
        discarded.

        This function emulates the behavior of `O_TMPFILE` for systems and
        file-systems where it is not available.

        Parameters:
        -----------
        rpath
            Relative path from cache-root to the location where to link the
            file on success.
        rpath_store
            Relative path from cache-root to the store to use for temporary
            files. This must share the same mount-instance as the final path.
        closefd
            If false, retain file-descriptor (and lock) on success.
        ignore_exist
            If true, an existing file at the desired location during a
            replacement will not cause an error.
        replace
            If true, replace a previous file at the specified location. If
            false, no replacement takes place and the temporary file is
            discarded.
        """

        assert not replace or not ignore_exist

        rpath_tmp = None

        try:
            # First create a random file in the selected store. This file will
            # have a UUID as name and thus we can safely use `O_CREAT|O_EXCL`
            # to create it and guarantee its uniqueness.
            name = "uuid-" + uuid.uuid4().hex
            rpath_tmp = os.path.join(rpath_store, name)
            with self._atomic_open(
                rpath_tmp,
                wait=True,
                write=True,
                closefd=closefd,
                oflags=os.O_CREAT | os.O_EXCL,
            ) as fd:
                # Yield control to the caller to fill in data and metadata.
                with os.fdopen(fd, "r+", closefd=False, encoding="utf8") as file:
                    yield file

                suppress = []
                if ignore_exist:
                    suppress.append(errno.EEXIST)

                if replace:
                    # Move the file into the desired location, possibly
                    # replacing any existing entry.
                    os.rename(
                        src=self._path(rpath_tmp),
                        dst=self._path(rpath),
                    )
                else:
                    # Preferably, we used `RENAME_NOREPLACE`, but this is not
                    # supported on NFS. Instead, we create a hard-link, which
                    # will fail if the target already exists. We rely on the
                    # cleanup-path to drop the original link.
                    with ctx.suppress_oserror(*suppress):
                        os.link(
                            src=self._path(rpath_tmp),
                            dst=self._path(rpath),
                            follow_symlinks=False,
                        )
        finally:
            if rpath_tmp is not None:
                # If the temporary file exists, we delete it. If we haven't
                # created it, or if we already moved it, this will be a no-op.
                # Due to the unique name, we will never delete a file we do not
                # own. If we hard-linked the file, this merely deletes the
                # original temporary link.
                # On fatal errors, we leak the file into the object store. Due
                # to the released lock and UUID name, cache management will
                # clean it up.
                with ctx.suppress_oserror(errno.ENOENT):
                    os.unlink(self._path(rpath_tmp))

    def _atomic_dir(self, rpath_store: str) -> Tuple[str, int]:
        """Atomically create and lock an anonymous directory

        Create an anonymous directory in the specified storage directory
        relative to the cache-root. The directory will have a UUID as name. On
        success, the name of the directory and the open file-descriptor to its
        acquired lock file (write-locked) are returned.

        The lock-file logic follows the cache-logic for objects. Hence, the
        cache scaffolding for the specified store must exist. No other cache
        infrastructure is required, though.

        Parameters:
        -----------
        rpath_store
            Relative path from the cache-root to the storage directory to create
            the new anonymous directory in. Most likely, this is either the
            object-store or the staging-area.
        """

        rpath_dir = None
        rpath_lock = None

        try:
            while True:
                # Allocate a UUID for the new directory and prepare the paths
                # to the directory and lock-file inside.
                name = "uuid-" + uuid.uuid4().hex
                rpath_dir = os.path.join(rpath_store, name)
                rpath_lock = os.path.join(rpath_dir, self._filename_object_lock)

                # Create an anonymous lock-file, but before linking it create
                # the target directory to link the file in. Use an ExitStack
                # to control exactly where to catch exceptions.
                with contextlib.ExitStack() as es:
                    f = es.enter_context(
                        self._atomic_file(
                            rpath_lock,
                            rpath_store,
                            closefd=False,
                        )
                    )
                    lockfd = f.fileno()
                    os.mkdir(self._path(rpath_dir))

                    # Exit the `_atomic_file()` context, thus triggering a link
                    # of the anonymous lock-file into the new directory. A
                    # parallel cleanup might have deleted the empty directory,
                    # so catch `ENOENT` and retry.
                    try:
                        es.close()
                    except OSError as e:
                        if e.errno == errno.ENOENT:
                            continue
                        raise

                return (name, lockfd)
        except BaseException:
            # On error, we might have already created the directory or even
            # linked the lock-file. Try unlinking both, but ignore errors if
            # they do not exist. Due to using UUIDs as names we cannot conflict
            # with entries created by some-one else.
            if rpath_lock is not None:
                with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
                    os.unlink(self._path(rpath_lock))
            if rpath_dir is not None:
                with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
                    os.rmdir(self._path(rpath_dir))
            raise

    def _create_scaffolding(self):
        """Create cache scaffolding

        Create the directories leading to the cache, as well as the internal
        scaffolding directories and files. This ensures that an existing cache
        is not interrupted or rewritten. Hence, this can safely be called in
        parallel, even on live caches.

        If this happens to create a new cache, it is initialized with its
        default configuration and constraints. By default, this means the cache
        has a maximum size of 0 and thus is only used as staging area with no
        long-time storage.

        This call requires no cache-infrastructure to be in place, and can be
        called repeatedly at any time.
        """

        # Create the directory-scaffolding of the cache. Make sure to ignore
        # errors when they already exist, to allow for parallel setups.
        dirs = [
            self._path(self._dirname_objects),
            self._path(self._dirname_stage),
        ]
        for i in dirs:
            os.makedirs(i, exist_ok=True)

        # Create the file-scaffolding of the cache. We fill in the default
        # information and ignore racing operations.
        with self._atomic_file(self._filename_cache_tag, self._dirname_objects, ignore_exist=True) as f:
            f.write(
                "Signature: 8a477f597d28d172789f06886806bc55\n"
                "# This is a cache directory tag created by osbuild (see https://bford.info/cachedir/)\n"
            )
        with self._atomic_file(self._filename_cache_info, self._dirname_objects, ignore_exist=True) as f:
            json.dump({"version": self._version_current}, f)
        with self._atomic_file(self._filename_cache_lock, self._dirname_objects, ignore_exist=True) as f:
            pass
        with self._atomic_file(self._filename_cache_size, self._dirname_objects, ignore_exist=True) as f:
            f.write("0")

    def _load_cache_info(self, info: Optional[FsCacheInfo] = None):
        """Load cache information

        This loads information about the cache into this cache-instance. The
        cache-information is itself cached on this instance and only updated
        on request. If the underlying file in the cache changes at runtime it
        is not automatically re-loaded. Only when this function is called the
        information is reloaded.

        By default this function reads the cache-information from the
        respective file in the cache and then caches it on this instance. If
        the `info` argument is not `None`, then no information is read from the
        file-system, but instead the information is taken from the `info`
        argument. This allows changing the cache-information of this instance
        without necessarily modifying the underlying file.

        This call requires the cache scaffolding to be fully created.

        Parameters:
        -----------
        info
            If `None`, the cache info file is read. Otherwise, the information
            is taken from this tuple.
        """

        # Parse the JSON data into python.
        if info is None:
            with open(self._path(self._filename_cache_info), "r", encoding="utf8") as f:
                info_raw = json.load(f)

            info = FsCacheInfo.from_json(info_raw)

        # Retain information.
        self._info = info

        # Parse `maximum-size` into internal representation.
        if info.maximum_size == "unlimited":
            self._info_maximum_size = -1
        elif isinstance(info.maximum_size, int):
            self._info_maximum_size = info.maximum_size
        elif info.maximum_size is None:
            self._info_maximum_size = 0
        else:
            raise ValueError(
                f"maximum-size can only be set to 'unlimited' or an integer value, got {type(info.maximum_size)}")

    def _is_active(self):
        # Internal helper to verify we are in an active context-manager.
        return self._active

    def _is_compatible(self):
        # Internal helper to verify the cache-version is supported.
        return self._info.version is not None and \
            self._version_minimum <= self._info.version <= self._version_current

    def __enter__(self):
        assert not self._active

        try:
            # Acquire the current boot-id so we can tag entries accordingly, and
            # judge entries that are from previous boots.
            self._bootid = linux.proc_boot_id(self._appid).hex

            # Create the scaffolding for the entire cache.
            self._create_scaffolding()

            # Acquire a shared cache lock.
            self._lock = os.open(
                self._path(self._filename_cache_lock),
                os.O_RDONLY | os.O_CLOEXEC,
            )
            linux.fcntl_flock(self._lock, linux.fcntl.F_RDLCK, wait=True)

            # Read the cache configuration.
            self._load_cache_info()

            self._active = True
            return self
        except BaseException:
            self.__exit__(None, None, None)
            raise

    def __exit__(self, exc_type, exc_value, exc_tb):
        # Discard any state of this context and reset to original state.
        if self._lock is not None:
            linux.fcntl_flock(self._lock, linux.fcntl.F_UNLCK)
            os.close(self._lock)
            self._lock = None
        self._active = False
        self._bootid = None
        self._info = FsCacheInfo()
        # We always have to leave the file-system scaffolding around. Even if
        # the cache is entirely empty, we cannot know whether there are other
        # parallel accesses (without unreasonable effort).

    def _update_cache_size(self, diff: int) -> bool:
        """Update cache size

        Update the total cache size by the specified amount, unless it exceeds
        the cache limits.

        This carefully updates the stored cache size to allow for parallel
        updates by other cache users. If the cache limits are exceeded, the
        operation is canceled and `False` is returned. Otherwise, `True` is
        returned.

        If the specified amount is negative, the operation always succeeds. If
        the cache size would end up negative, it is capped at 0.

        This operation requires an active context.
        """

        assert self._is_active()
        assert self._is_compatible()

        # Open the cache-size and lock it for writing. But instead of writing
        # directly to it, we replace it with a new file. This guarantees that
        # we cannot crash while writing a partial size, but always atomically
        # update the content.
        with self._atomic_open(self._filename_cache_size, write=True, wait=True) as fd:
            with os.fdopen(fd, "r", closefd=False, encoding="utf8") as f:
                size = json.load(f)

            if size + diff < 0:
                size = 0
            elif (self._info_maximum_size < 0) or (size + diff <= self._info_maximum_size):
                size = size + diff
            else:
                return False

            with self._atomic_file(self._filename_cache_size, self._dirname_objects, replace=True) as f:
                json.dump(size, f)

            return True

    def _rm_r_object(self, rpath_dir: str):
        """Remove object

        Recursively remove all traces of a stored object. This either requires
        the caller to hold a write-lock on the entry, or otherwise guarantee
        that no cache lookups can acquire the entry concurrently.

        This carefully deletes any traces of the entry, making sure to first
        mark the object as invalid, and dropping the lock-file last. This can
        safely be called on partially constructured or non-existing entries.

        Parameters:
        -----------
        rpath_dir
            Relative path from the cache-root to the object directory.
        """

        path_dir = self._path(rpath_dir)
        path_info = os.path.join(path_dir, self._filename_object_info)
        path_lock = os.path.join(path_dir, self._filename_object_lock)

        # Optimization: Bail out early if the entry is non-existant
        if not os.path.lexists(path_dir):
            return

        # First step, we unlink the info-file. This will mark the entry as
        # volatile and thus it will get cleaned up by cache management in case
        # we crash while deleting it. Furthermore, no cache lookups will ever
        # consider the entry again if the info-file is missing.
        with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
            os.unlink(path_info)

        # Now iterate the directory and drop everything _except_ the lock file.
        # This makes sure no parallel operation will needlessly race with us. In
        # case no lock is acquired, we still allow for parallel racing cleanups.
        #
        # Note that racing cleanups might delete the entire directory at any
        # time during this iteration. Furthermore, `scandir()` is not atomic but
        # repeatedly calls into the kernel. Hence, we carefully bail out once
        # it reports a non-existant directory.
        with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
            for entry in os.scandir(path_dir):
                if entry.name == self._filename_object_lock:
                    continue
                with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
                    if entry.is_dir():
                        rmrf.rmtree(entry.path)
                    else:
                        os.unlink(entry.path)

        # With everything gone, we unlink the lock-file and eventually delete
        # the directory. Again, cleanup routines might have raced us, so avoid
        # failing in case the entries are already gone.
        with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
            os.unlink(path_lock)
        with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
            os.rmdir(path_dir)

    @contextlib.contextmanager
    def stage(self):
        """Create staging entry

        Create a new entry in the staging area and yield control to the caller
        with the relative path to the entry. Once control returns, the staging
        entry is completely discarded.

        If the application crashes while holding a staging entry, it will be
        left behind in the staging directory, but unlocked and marked as stale.
        Hence, any cache management routine will discard it.
        """

        # We check for an active context, but we never check for
        # version-compatibility, because there is no way we can run without
        # a staging area. Hence, the staging-area has to be backwards
        # compatible at all times.
        assert self._is_active()

        uuidname = None
        lockfd = None

        try:
            # Create and lock a new anonymous object in the staging area.
            uuidname, lockfd = self._atomic_dir(self._dirname_stage)

            rpath_data = os.path.join(
                self._dirname_stage,
                uuidname,
                self._dirname_data,
            )

            # Prepare an empty data directory and yield it to the caller.
            os.mkdir(self._path(rpath_data))
            yield rpath_data
        finally:
            if lockfd is not None:
                self._rm_r_object(os.path.join(self._dirname_stage, uuidname))
                linux.fcntl_flock(lockfd, linux.fcntl.F_UNLCK)
                os.close(lockfd)

    @contextlib.contextmanager
    def store(self, name: str):
        """Store object in cache

        Create a new entry and store it in the cache with the specified name.
        The entry is first created with an anonymous name and control is yielded
        to the caller to fill in data. Once control returns, the entry is
        committed with the specified name.

        The final commit is skipped if an entry with the given name already
        exists, or its name is claimed for other reasons. Furthermore, the
        commit is skipped if cache limits are exceeded, or if cache maintenance
        refuses the commit. Hence, a commit can never be relied upon and the
        entry might be deleted from the cache as soon as the commit was invoked.

        Parameters:
        -----------
        name
            Name to store the object under.
        """

        assert self._is_active()
        assert self._bootid is not None

        if not name:
            raise ValueError()

        # If the cache-version is incompatible to this implementation, we short
        # this call into the staging-area (which is always compatible). This
        # avoids raising an exception (at the cost of dealing with this in the
        # caller), and instead just creates a temporary copy which we discard.
        if not self._is_compatible():
            with self.stage() as p:
                yield p
            return

        uuidname = None
        lockfd = None

        try:
            # Create and lock a new anonymous object in the staging area.
            uuidname, lockfd = self._atomic_dir(self._dirname_objects)

            rpath_uuid = os.path.join(
                self._dirname_objects,
                uuidname,
            )
            rpath_data = os.path.join(
                rpath_uuid,
                self._dirname_data,
            )
            rpath_info = os.path.join(
                rpath_uuid,
                self._filename_object_info,
            )
            path_uuid = self._path(rpath_uuid)
            path_data = self._path(rpath_data)
            path_info = self._path(rpath_info)

            # Prepare an empty data directory and yield it to the caller.
            os.mkdir(path_data)
            yield rpath_data

            # Collect metadata about the new entry.
            info: Dict[str, Any] = {}
            info["creation-boot-id"] = self._bootid
            info["size"] = self._calculate_space(path_data)

            # Exit early if it never is going to fit
            if self._info_maximum_size > -1 and info["size"] > self._info_maximum_size:
                return

            # Update the total cache-size. If it exceeds the limits, remove
            # least recently used objects until there is enough space.
            #
            # Note that if we crash after updating the total cache size, but
            # before committing the object information, the total cache size
            # will be out of sync.
            #
            # However, it is never overcommitted, so we will never
            # violate any cache invariants. Future code needs to resync
            # the cache (e.g. on open with some simple journal strategy).
            if not self._update_cache_size(info["size"]):
                # try to free space
                self._remove_lru(info["size"])
                # and see if the update can happen now
                if not self._update_cache_size(info["size"]):
                    # stil could not free enough space
                    return

            try:
                # Commit the object-information, thus marking it as fully
                # committed and accounted in the cache.
                with open(path_info, "x", encoding="utf8") as f:
                    json.dump(info, f)

                # As last step move the entry to the desired location. If the
                # target name is already taken, we bail out and pretend the
                # entry was immediately overwritten by another one.
                #
                # Preferably, we used RENAME_NOREPLACE, but this is not
                # available on all file-systems. Hence, we rely on the fact
                # that non-empty directories cannot be replaced, so we
                # automatically get RENAME_NOREPLACE behavior.
                path_name = self._path(self._dirname_objects, name)
                try:
                    os.rename(
                        src=path_uuid,
                        dst=path_name,
                    )
                except OSError as e:
                    ignore = [errno.EEXIST, errno.ENOTDIR, errno.ENOTEMPTY]
                    if e.errno not in ignore:
                        raise

                uuidname = None
            finally:
                # If the anonymous entry still exists, it will be cleaned up by
                # the outer handler. Hence, make sure to drop the info file
                # again and de-account it, so we don't overcommit.
                if os.path.lexists(path_uuid):
                    with ctx.suppress_oserror(errno.ENOENT, errno.ENOTDIR):
                        os.unlink(path_info)
                    self._update_cache_size(-info["size"])
        finally:
            if lockfd is not None:
                if uuidname is not None:
                    # In case this runs after the object was renamed, but before
                    # `uuidname` was cleared, then `_rm_r_object()` will be a
                    # no-op.
                    self._rm_r_object(os.path.join(self._dirname_objects, uuidname))
                linux.fcntl_flock(lockfd, linux.fcntl.F_UNLCK)
                os.close(lockfd)

    @contextlib.contextmanager
    def load(self, name: str):
        """Load a cache entry

        Find the cache entry with the given name, acquire a read-lock and
        yield its path back to the caller. Once control returns, the entry
        is released.

        The returned path is the relative path between the cache and the top
        level directory of the cache entry.

        Parameters:
        -----------
        name
            Name of the cache entry to find.
        """

        assert self._is_active()

        if not name:
            raise ValueError()
        if not self._is_compatible():
            raise self.MissError()

        with contextlib.ExitStack() as es:
            # Use an ExitStack so we can catch exceptions raised by the
            # `__enter__()` call on the context-manager. We want to catch
            # `OSError` exceptions and convert them to cache-misses.
            obj_lock_path = os.path.join(
                self._dirname_objects, name, self._filename_object_lock)
            try:
                lock_fd = es.enter_context(
                    self._atomic_open(
                        obj_lock_path,
                        write=False,
                        wait=False,
                    )
                )
            except OSError as e:
                if e.errno in [errno.EAGAIN, errno.ENOENT, errno.ENOTDIR]:
                    raise self.MissError() from None
                raise e

            libc = linux.Libc.default()
            libc.futimens(lock_fd, ctypes.byref(linux.c_timespec_times2(
                atime=linux.c_timespec(tv_sec=0, tv_nsec=libc.UTIME_NOW),
                mtime=linux.c_timespec(tv_sec=0, tv_nsec=libc.UTIME_OMIT),
            )))

            yield os.path.join(
                self._dirname_objects,
                name,
                self._dirname_data,
            )

    def _last_used(self, name: str) -> float:
        """Return the last time the given object was last used.

        Note that the resolution is only as good as what the filesystem "atime"
        gives us.
        """
        obj_lock_path = os.path.join(
            self._dirname_objects, name, self._filename_object_lock)
        try:
            return os.stat(self._path(obj_lock_path)).st_atime
        except OSError as e:
            if e.errno in [errno.EAGAIN, errno.ENOENT, errno.ENOTDIR]:
                raise self.MissError() from None
            raise e

    def _last_used_objs(self) -> List[FsCacheObjectInfo]:
        """Return a list of FsCacheObjectInfo with name, last_used
           information sorted by last_used time.

        Note that this function will be racy when used without a lock and
        the caller needs to handle this.
        """
        objs = []
        for name in os.listdir(self._path(self._dirname_objects)):
            try:
                last_used = self._last_used(name)
            except (OSError, FsCache.MissError):
                continue
            objs.append(FsCacheObjectInfo(name=name, last_used=last_used))
        return sorted(objs, key=lambda obj: obj.last_used)

    def _remove_lru(self, required_size):
        """"
        Make room in the cache for "required_size" by remove the least
        recently used entry from the cache. Note that the cache may
        clear more than required_size.
        """
        # To avoid having to take a global cache lock the strategy is:
        # 1. Get list of (object, last_used) sorted from oldest to newest.
        #    This is racy so we need to take care of that in step(2).
        # 2. Start with the oldest entry, try to take a write_lock
        #    (with O_NOATIME to be extra sure that atime information is
        #    correct). Get the "last_used" (atime) time and compare to what
        #    we expect in the list. If it diverges the object got load()ed
        #    while we iterated. Skip it and go to (2) again.
        # 3. Remove entry, update cache size after the entry is removed.
        #
        # Note that there is a risk to get out-of-sync in (3). If the
        # process dies while removing and before updating the cache
        # size the cache will be over reported.

        # Try to clean at least twice the requested size to avoid having
        # to do this all over again
        try_to_free = required_size * 2
        freed_so_far = 0
        for name, last_used in self._last_used_objs():
            # take write lock for the indivdual object
            rpath = os.path.join(self._dirname_objects, name)
            rpath_lock = os.path.join(rpath, self._filename_object_lock)
            # Ideally there would some lock helper instead of the low-level
            # file manipulation to abstract this a bit more.
            try:
                with self._atomic_open(
                        rpath_lock,
                        wait=False,
                        write=True,
                        # atime carries the "last-used" data so don't alter it
                        oflags=os.O_EXCL | os.O_NOATIME,
                ):
                    if last_used != self._last_used(name):
                        continue
                    # This is racy right now if the process is killed
                    # during "_rm_r_object(rpath)" because then the
                    # cache size is never reduced by the amount that
                    # was about to be deleted.
                    #
                    # To fix it we need to (atomic) rename the
                    # "object.info" file in _rm_r_object() to
                    # something like "object.removing". Then when
                    # opening the cache scan for leftover
                    # "object.removing" files and finish the cleanup
                    # and update the cache size based on the size
                    # recorded inside "object.removing".
                    size = self._calculate_space(self._path(rpath))
                    self._rm_r_object(rpath)
                    self._update_cache_size(-size)
                    freed_so_far += size
                    if freed_so_far >= try_to_free:
                        break
            except BlockingIOError:
                continue
        # return True if at least the required size got freed
        return freed_so_far > required_size

    @property
    def info(self) -> FsCacheInfo:
        """Query Cache Information

        Return the parsed cache information which is currently cached on this
        cache-instance. The cache information has all unknown fields stripped.

        Unset values are represented by `None`, and the cache will interpret
        it as the default value for the respective field.
        """

        assert self._is_active()

        return self._info

    @info.setter
    def info(self, info: FsCacheInfo):
        """Write Cache Information

        Update and write the cache-information onto the file-system. This first
        locks the cache-information file, reads it in, updates the newly read
        information with the data from `info`, writes the result back to disk
        and finally unlocks the file.

        There are a few caveats to take into account:

         * The locking guarantees that simultaneous updates will be properly
           ordered and never discard any information.
         * Since this reads in the newest cache-information, this function can
           update cache-information values other than the ones from `info`. Any
           value unset in `info` will be re-read from disk and thus might
           change (in the future, if required, this can be adjusted to allow a
           caller to hook into the operation while the lock is held).
         * You cannot strip known values from the cache-information. Any value
           not present in `info` is left unchanged. You must explicitly set a
           value to its default to reset it.
         * Cache-information fields that are not known to this implementation
           are never exposed to the caller, but are left unchanged on-disk.
           This guarantees that future extensions are left alone and are not
           accidentally stripped.

        The cached information of this instance is updated to reflect the
        changes.

        Parameters:
        -----------
        info
            Cache information object to consume and write.
        """

        assert self._is_active()

        with self._atomic_open(self._filename_cache_info, write=True, wait=True) as fd:
            with os.fdopen(fd, "r", closefd=False, encoding="utf8") as f:
                info_raw = json.load(f)

            # If the on-disk data is in an unexpected format, we never touch
            # it. If it is a JSON-object, we update it with the new values and
            # then re-parse it into a full `FsCacheInfo` with all known fields
            # populated.
            if isinstance(info_raw, dict):
                info_raw.update(info.to_json())
                info = FsCacheInfo.from_json(info_raw)

                # Replace the file with the new values. This releases the lock.
                if self._is_compatible():
                    with self._atomic_file(self._filename_cache_info, self._dirname_objects, replace=True) as f:
                        json.dump(info_raw, f)

        self._load_cache_info(info)

    def store_tree(self, name: str, tree: Any):
        """Store file system tree in cache

        Create a new entry in the object store containing a copy of the file
        system tree specified as `tree`. This behaves like `store()` but instead
        of providing a context to the caller it will copy the specified tree.

        Similar to `store()`, when the entry is committed it is immediately
        unlocked and released to the cache. This means it might vanish at any
        moment due to a parallel cleanup. Hence, a caller cannot rely on the
        object being available in the cache once this call returns.

        If `tree` points to a file, the file is copied. If it points to a
        directory, the entire directory tree is copied including the root entry
        itself. To copy an entire directory without its root entry, use the
        `path/.` notation. Links are never followed but copied verbatim.
        All metadata is preserved, if possible.

        Parameters:
        -----------
        name
            Name to store the object under.
        tree:
            Path to the file system tree to copy.
        """

        with self.store(name) as rpath_data:
            r = subprocess.run(
                [
                    "cp",
                    "--reflink=auto",
                    "-a",
                    "--",
                    os.fspath(tree),
                    self._path(rpath_data),
                ],
                check=False,
                encoding="utf-8",
                stderr=subprocess.STDOUT,
                stdout=subprocess.PIPE,
            )
            if r.returncode != 0:
                code = r.returncode
                msg = r.stdout.strip()
                raise RuntimeError(f"Cannot copy into file-system cache ({code}): {msg}")



#!/usr/bin/python3

import os.path
from unittest.mock import call, patch

import pytest  # type: ignore

from osbuild import testutil

STAGE_NAME = "org.osbuild.selinux"


def get_test_input(test_data, file_contexts=False, labels=False):
    test_input = {
        "type": STAGE_NAME,
        "options": {}
    }
    if file_contexts:
        test_input["options"]["file_contexts"] = "some-context"
    if labels:
        test_input["options"]["labels"] = {"/path/to/file": "label_to_apply"}

    test_input["options"].update(test_data)
    return test_input


@pytest.mark.parametrize("test_data,expected_err", [
    # good
    ({"labels": {"/usr/bin/cp": "system_u:object_r:install_exec_t:s0"}}, ""),
    ({"force_autorelabel": True}, ""),
    ({"exclude_paths": ["/sysroot"]}, ""),
    ({"target": "mount://disk/boot/efi"}, ""),
    ({"target": "tree:///boot/efi"}, ""),
    ({"file_contexts": "path/to/file_contexts"}, ""),
    # bad
    ({"target": "/boot/efi"}, "'/boot/efi' does not match '^mount://[^/]+/|^tree:///'"),
    ({"file_contexts": 1234}, "1234 is not of type 'string'"),
    ({"labels": "xxx"}, "'xxx' is not of type 'object'"),
    ({"force_autorelabel": "foo"}, "'foo' is not of type 'boolean'"),
])
def test_schema_validation_selinux(stage_schema, test_data, expected_err):
    res = stage_schema.validate(get_test_input(test_data, file_contexts=True, labels=False))
    if expected_err == "":
        assert res.valid is True, f"err: {[e.as_dict() for e in res.errors]}"
    else:
        assert res.valid is False
        testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)


def test_schema_validation_selinux_required_options(stage_schema):
    res = stage_schema.validate(get_test_input({}, file_contexts=False, labels=False))
    assert res.valid is False
    expected_err = "{} is not valid under any of the given schemas"
    testutil.assert_jsonschema_error_contains(res, expected_err, expected_num_errs=1)
    res = stage_schema.validate(get_test_input({}, file_contexts=True, labels=False))
    assert res.valid is True
    res = stage_schema.validate(get_test_input({}, file_contexts=False, labels=True))
    assert res.valid is True


@patch("osbuild.util.selinux.setfiles")
def test_selinux_file_contexts_schemeless_relpath(mocked_setfiles, tmp_path, stage_module):
    tree = tmp_path / "tree"

    args = {
        "tree": tree,
        "options": {
            "file_contexts": "etc/selinux/thing",
        }
    }
    stage_module.main(args)

    assert len(mocked_setfiles.call_args_list) == 1
    args, kwargs = mocked_setfiles.call_args_list[0]
    assert args == (f"{tree}/etc/selinux/thing", os.fspath(tree), "/")
    assert kwargs == {"exclude_paths": None}


@patch("osbuild.util.selinux.setfiles")
def test_selinux_file_contexts_schemeless_abspath(mocked_setfiles, tmp_path, stage_module):
    tree = tmp_path / "tree"

    args = {
        "tree": tree,
        "options": {
            "file_contexts": "/etc/selinux/thing",
        }
    }
    stage_module.main(args)

    assert len(mocked_setfiles.call_args_list) == 1
    args, kwargs = mocked_setfiles.call_args_list[0]
    assert args == (f"{tree}/etc/selinux/thing", os.fspath(tree), "/")
    assert kwargs == {"exclude_paths": None}


@patch("osbuild.util.selinux.setfiles")
def test_selinux_file_contexts_mounts(mocked_setfiles, tmp_path, stage_module):
    tree = tmp_path / "tree"
    mounts = tmp_path / "mounts"
    root = mounts / "."

    args = {
        "tree": tree,
        "options": {
            "file_contexts": "mount://root/etc/selinux/file_contexts",
            "target": "mount://root/"
        },
        "paths": {
            "mounts": mounts,
        },
        "mounts": {
            "root": {
                "path": root
            }
        }
    }
    stage_module.main(args)

    assert len(mocked_setfiles.call_args_list) == 1
    args, kwargs = mocked_setfiles.call_args_list[0]
    assert args == (f"{mounts}/etc/selinux/file_contexts", f"{mounts}", "/")
    assert kwargs == {"exclude_paths": None}


@patch("osbuild.util.selinux.setfiles")
def test_selinux_file_contexts_exclude(mocked_setfiles, tmp_path, stage_module):
    tree = tmp_path / "tree"

    args = {
        "tree": tree,
        "options": {
            "file_contexts": "etc/selinux/thing",
            "exclude_paths": ["/sysroot"],
        }
    }
    stage_module.main(args)

    assert len(mocked_setfiles.call_args_list) == 1
    args, kwargs = mocked_setfiles.call_args_list[0]
    assert args == (f"{tree}/etc/selinux/thing", os.fspath(tree), "/")
    assert kwargs == {"exclude_paths": [f"{tree}/sysroot"]}


@patch("osbuild.util.selinux.setfilecon")
@patch("osbuild.util.selinux.setfiles")
def test_selinux_labels(mocked_setfiles, mocked_setfilecon, tmp_path, stage_module):
    tree = tmp_path / "tree"
    testutil.make_fake_input_tree(tmp_path, {
        "/usr/bin/echo": "I'm only an imposter",
        "/sbin/sulogin": "I'm only an imposter",
    })

    options = {
        "file_contexts": "etc/selinux/thing",
        "labels": {
            "tree:///usr/bin/echo": "system_u:object_r:bin_t:s0",
            "/sbin/sulogin": "system_u:object_r:sulogin_exec_t:s0",
        }
    }
    args = {
        "tree": tree,
        "options": options
    }
    stage_module.main(args)

    assert len(mocked_setfiles.call_args_list) == 1
    assert len(mocked_setfilecon.call_args_list) == 2
    assert mocked_setfilecon.call_args_list == [
        call(f"{tree}/usr/bin/echo", "system_u:object_r:bin_t:s0"),
        call(f"{tree}/sbin/sulogin", "system_u:object_r:sulogin_exec_t:s0"),
    ]


@patch("osbuild.util.selinux.setfilecon")
def test_selinux_labels_mount(mocked_setfilecon, tmp_path, stage_module):
    tree = tmp_path / "tree"
    mounts = tmp_path / "mounts"

    testutil.make_fake_tree(mounts, {"/sbin/su": "I'm only an imposter"})
    args = {
        "tree": tree,
        "options": {
            "labels": {
                "mount://root/sbin/su": "system_u:object_r:su_exec_t:s0",
            }
        },
        "paths": {
            "mounts": mounts,
        },
        "mounts": {
            "root": {"path": mounts}
        }
    }
    stage_module.main(args)

    assert len(mocked_setfilecon.call_args_list) == 1
    assert mocked_setfilecon.call_args_list == [
        call(f"{mounts}/sbin/su", "system_u:object_r:su_exec_t:s0"),
    ]


@patch("osbuild.util.selinux.setfiles")
def test_selinux_force_autorelabel(mocked_setfiles, tmp_path, stage_module):  # pylint: disable=unused-argument
    for enable_autorelabel in [False, True]:
        options = {
            "file_contexts": "etc/selinux/thing",
            "force_autorelabel": enable_autorelabel,
        }
        args = {
            "tree": f"{tmp_path}",
            "options": options
        }
        stage_module.main(args)

        assert (tmp_path / ".autorelabel").exists() == enable_autorelabel
        if enable_autorelabel:
            assert (tmp_path / ".autorelabel").read_text() == "-F"



#!/usr/bin/python3

import os.path

import pytest

STAGE_NAME = "org.osbuild.grub2.legacy"


# Test that the /etc/default/grub file is created with the correct content
@pytest.mark.parametrize("test_data,expected_conf", [
    # default
    ({}, """GRUB_TIMEOUT=0
GRUB_CMDLINE_LINUX=""
GRUB_DISABLE_SUBMENU=true
GRUB_DISABLE_RECOVERY=true
GRUB_TIMEOUT_STYLE=countdown
GRUB_DEFAULT=saved
"""),
    # custom
    ({
        "default": "0",
        "disable_submenu": False,
        "disable_recovery": False,
        "timeout": 10,
        "timeout_style": "hidden",
    }, """GRUB_TIMEOUT=10
GRUB_CMDLINE_LINUX=""
GRUB_DISABLE_SUBMENU=false
GRUB_DISABLE_RECOVERY=false
GRUB_TIMEOUT_STYLE=hidden
GRUB_DEFAULT=0
"""),
    # custom (Azure)
    ({
        "cmdline": "loglevel=3 crashkernel=auto console=tty1 console=ttyS0 earlyprintk=ttyS0 rootdelay=300",
        "disable_submenu": True,
        "disable_recovery": True,
        "distributor": "$(sed 's, release .*$,,g' /etc/system-release)",
        "serial": "serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1",
        "terminal": ["serial", "console"],
        "terminal_output": ["console"],
        "timeout": 10,
        "timeout_style": "countdown",
    }, """GRUB_TIMEOUT=10
GRUB_CMDLINE_LINUX="loglevel=3 crashkernel=auto console=tty1 console=ttyS0 earlyprintk=ttyS0 rootdelay=300"
GRUB_DISABLE_SUBMENU=true
GRUB_DISABLE_RECOVERY=true
GRUB_TIMEOUT_STYLE=countdown
GRUB_DEFAULT=saved
GRUB_DISTRIBUTOR="$(sed 's, release .*$,,g' /etc/system-release)"
GRUB_SERIAL_COMMAND="serial --speed=115200 --unit=0 --word=8 --parity=no --stop=1"
GRUB_TERMINAL="serial console"
GRUB_TERMINAL_OUTPUT="console"
"""),
])
def test_grub2_default_conf(tmp_path, stage_module, test_data, expected_conf):
    treedir = tmp_path / "tree"
    confpath = treedir / "etc/default/grub"
    confpath.parent.mkdir(parents=True, exist_ok=True)

    options = {
        "rootfs": {
            "label": "root"
        },
        "entries": [
            {
                "id": "fff",
                "kernel": "4.18",
                "product": {
                    "name": "Fedora",
                    "version": "40"
                }
            }
        ]
    }

    options["config"] = test_data
    stage_module.main(treedir, options)

    assert os.path.exists(confpath)
    assert confpath.read_text() == expected_conf
